# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Selection of quantum circuits architectures used in defining QHBMs."""

import bisect

import cirq
import sympy
import tensorflow_quantum as tfq

from qhbmlib import ebm

# ============================================================================ #
# HEA components.
# ============================================================================ #


def get_xz_rotation(q, a, b):
  """Two-axis single qubit rotation."""
  return cirq.Circuit(cirq.X(q)**a, cirq.Z(q)**b)


def get_xyz_rotation(q, a, b, c):
  """General single qubit rotation."""
  return cirq.Circuit(cirq.X(q)**a, cirq.Y(q)**b, cirq.Z(q)**c)


def get_cz_exp(q0, q1, a):
  """Exponent of entangling CZ gate."""
  return cirq.Circuit(cirq.CZPowGate(exponent=a)(q0, q1))


def get_zz_exp(q0, q1, a):
  """Exponent of entangling ZZ gate."""
  return cirq.Circuit(cirq.ZZPowGate(exponent=a)(q0, q1))


def get_cnot_exp(q0, q1, a):
  """Exponent of entangling CNot gate."""
  return cirq.Circuit(cirq.CNotPowGate(exponent=a)(q0, q1))


def get_xz_rotation_layer(qubits, layer_num, name):
  """Apply two-axis single qubit rotations to all the given qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for n, q in enumerate(qubits):
    sx, sz = sympy.symbols("sx_{0}_{1}_{2} sz_{0}_{1}_{2}".format(
        name, layer_num, n))
    layer_symbols += [sx, sz]
    circuit += get_xz_rotation(q, sx, sz)
  return circuit, layer_symbols


def get_cz_exp_layer(qubits, layer_num, name):
  """Apply parameterized CZ gates to all pairs of nearest-neighbor qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for n, (q0, q1) in enumerate(zip(qubits[::2], qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n))
    layer_symbols += [a]
    circuit += get_cz_exp(q0, q1, a)
  shifted_qubits = qubits[1::]
  for n, (q0, q1) in enumerate(zip(shifted_qubits[::2], shifted_qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n + 1))
    layer_symbols += [a]
    circuit += get_cz_exp(q0, q1, a)
  return circuit, layer_symbols


def get_hardware_efficient_model_unitary(qubits, num_layers, name):
  """Build our full parameterized model unitary."""
  circuit = cirq.Circuit()
  all_symbols = []
  for layer_num in range(num_layers):
    new_circ, new_symb = get_xz_rotation_layer(qubits, layer_num, name)
    circuit += new_circ
    all_symbols += new_symb
    if len(qubits) > 1:
      new_circ, new_symb = get_cz_exp_layer(qubits, layer_num, name)
      circuit += new_circ
      all_symbols += new_symb
  return circuit, all_symbols


def get_cnot_exp_layer(qubits, layer_num, name):
  """Apply CNot gates to all pairs of nearest-neighbor qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for n, (q0, q1) in enumerate(zip(qubits[::2], qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n))
    layer_symbols += [a]
    circuit += get_cnot_exp(q0, q1, a)
  shifted_qubits = qubits[1::]
  for n, (q0, q1) in enumerate(zip(shifted_qubits[::2], shifted_qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n + 1))
    layer_symbols += [a]
    circuit += get_cnot_exp(q0, q1, a)
  return circuit, layer_symbols


def hea_1d_cnot(qubits, num_layers, name):
  """Build our full parameterized model unitary."""
  circuit = cirq.Circuit()
  all_symbols = []
  for layer_num in range(num_layers):
    new_circ, new_symb = get_xz_rotation_layer(qubits, layer_num, name)
    circuit += new_circ
    all_symbols += new_symb
    if len(qubits) > 1:
      new_circ, new_symb = get_cnot_exp_layer(qubits, layer_num, name)
      circuit += new_circ
      all_symbols += new_symb
  return circuit, all_symbols


def get_zz_exp_layer(qubits, layer_num, name):
  """Apply ZZ gates to all pairs of nearest-neighbor qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for n, (q0, q1) in enumerate(zip(qubits[::2], qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n))
    layer_symbols += [a]
    circuit += get_zz_exp(q0, q1, a)
  shifted_qubits = qubits[1::]
  for n, (q0, q1) in enumerate(zip(shifted_qubits[::2], shifted_qubits[1::2])):
    a = sympy.symbols("sc_{0}_{1}_{2}".format(name, layer_num, 2 * n + 1))
    layer_symbols += [a]
    circuit += get_zz_exp(q0, q1, a)
  return circuit, layer_symbols


def hea_1d_zz(qubits, num_layers, name):
  """Build our full parameterized model unitary."""
  circuit = cirq.Circuit()
  all_symbols = []
  for layer_num in range(num_layers):
    new_circ, new_symb = get_xz_rotation_layer(qubits, layer_num, name)
    circuit += new_circ
    all_symbols += new_symb
    if len(qubits) > 1:
      new_circ, new_symb = get_zz_exp_layer(qubits, layer_num, name)
      circuit += new_circ
      all_symbols += new_symb
  return circuit, all_symbols


# ============================================================================ #
# 2D HEA.
# ============================================================================ #


def get_2d_xz_rotation_layer(rows, cols, layer_num, name):
  """Apply single qubit rotations on a grid of qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for r in range(rows):
    for c in range(cols):
      sx = sympy.Symbol(f"sx_{name}_{layer_num}_{r}_{c}")
      sz = sympy.Symbol(f"sz_{name}_{layer_num}_{r}_{c}")
      layer_symbols += [sx, sz]
      circuit += get_xz_rotation(cirq.GridQubit(r, c), sx, sz)
  return circuit, layer_symbols


def get_2d_cz_exp_layer(rows, cols, layer_num, name):
  """Apply CZ gates to all pairs of nearest-neighbor qubits on a grid."""
  layer_symbols = []
  circuit = cirq.Circuit()
  # Apply horizontal bonds
  for r in range(rows):
    for par in [0, 1]:
      for q_c_0, q_c_1 in zip(range(par, cols, 2), range(par + 1, cols, 2)):
        scz = sympy.Symbol(f"scz_{name}_{layer_num}_row{r}_{q_c_0}_{q_c_1}")
        layer_symbols += [scz]
        circuit += get_cz_exp(
            cirq.GridQubit(r, q_c_0), cirq.GridQubit(r, q_c_1), scz)
  # Apply vertical bonds
  for c in range(cols):
    for par in [0, 1]:
      for q_r_0, q_r_1 in zip(range(par, rows, 2), range(par + 1, rows, 2)):
        scz = sympy.Symbol(f"scz_{name}_{layer_num}_col{c}_{q_r_0}_{q_r_1}")
        layer_symbols += [scz]
        circuit += get_cz_exp(
            cirq.GridQubit(q_r_0, c), cirq.GridQubit(q_r_1, c), scz)
  return circuit, layer_symbols


def get_2d_hea(rows, cols, num_layers, name):
  """Build a 2D HEA ansatz.

    Args:
      rows: int specifying the number of rows in the ansatz.
      cols: int specifying the number of columns in the ansatz.
      num_layers: int specifying how many layers of 2D HEA to apply.
      name: string which will be included in the parameters of the ansatz.

    Returns:
      circuit: `cirq.Circuit` which is the ansatz.
      symbols: list of `sympy.Symbol`s which are the parameters of the model.
    """
  symbols = []
  circuit = cirq.Circuit()
  for layer in range(num_layers):
    xz_circuit, xz_symbols = get_2d_xz_rotation_layer(rows, cols, layer, name)
    circuit += xz_circuit
    symbols += xz_symbols
    cz_circuit, cz_symbols = get_2d_cz_exp_layer(rows, cols, layer, name)
    circuit += cz_circuit
    symbols += cz_symbols
  return circuit, symbols


def get_2d_xyz_rotation_layer(rows, cols, layer_num, name):
  """Apply single qubit rotations on a grid of qubits."""
  layer_symbols = []
  circuit = cirq.Circuit()
  for r in range(rows):
    for c in range(cols):
      sx = sympy.Symbol(f"sx_{name}_{layer_num}_{r}_{c}")
      sy = sympy.Symbol(f"sy_{name}_{layer_num}_{r}_{c}")
      sz = sympy.Symbol(f"sz_{name}_{layer_num}_{r}_{c}")
      layer_symbols += [sx, sy, sz]
      circuit += get_xyz_rotation(cirq.GridQubit(r, c), sx, sy, sz)
  return circuit, layer_symbols


def get_2d_hea_y(rows, cols, num_layers, name):
  """Build a 2D HEA ansatz.

    Args:
      rows: int specifying the number of rows in the ansatz.
      cols: int specifying the number of columns in the ansatz.
      num_layers: int specifying how many layers of 2D HEA to apply.
      name: string which will be included in the parameters of the ansatz.

    Returns:
      circuit: `cirq.Circuit` which is the ansatz.
      symbols: list of `sympy.Symbol`s which are the parameters of the model.
    """
  symbols = []
  circuit = cirq.Circuit()
  for layer in range(num_layers):
    xyz_circuit, xyz_symbols = get_2d_xyz_rotation_layer(
        rows, cols, layer, name)
    circuit += xyz_circuit
    symbols += xyz_symbols
    cz_circuit, cz_symbols = get_2d_cz_exp_layer(rows, cols, layer, name)
    circuit += cz_circuit
    symbols += cz_symbols
  return circuit, symbols


def get_2d_cnot_exp_layer(rows, cols, layer_num, name):
  """Apply CNot gates to all pairs of nearest-neighbor qubits on a grid."""
  layer_symbols = []
  circuit = cirq.Circuit()
  # Apply horizontal bonds
  for r in range(rows):
    for par in [0, 1]:
      for q_c_0, q_c_1 in zip(range(par, cols, 2), range(par + 1, cols, 2)):
        scnot = sympy.Symbol(f"scnot_{name}_{layer_num}_row{r}_{q_c_0}_{q_c_1}")
        layer_symbols += [scnot]
        circuit += get_cnot_exp(
            cirq.GridQubit(r, q_c_0), cirq.GridQubit(r, q_c_1), scnot)
  # Apply vertical bonds
  for c in range(cols):
    for par in [0, 1]:
      for q_r_0, q_r_1 in zip(range(par, rows, 2), range(par + 1, rows, 2)):
        scnot = sympy.Symbol(f"scnot_{name}_{layer_num}_col{c}_{q_r_0}_{q_r_1}")
        layer_symbols += [scnot]
        circuit += get_cnot_exp(
            cirq.GridQubit(q_r_0, c), cirq.GridQubit(q_r_1, c), scnot)
  return circuit, layer_symbols


def get_2d_hea_cnot(rows, cols, num_layers, name):
  """Build a 2D HEA ansatz.

    Args:
      rows: int specifying the number of rows in the ansatz.
      cols: int specifying the number of columns in the ansatz.
      num_layers: int specifying how many layers of 2D HEA to apply.
      name: string which will be included in the parameters of the ansatz.

    Returns:
      circuit: `cirq.Circuit` which is the ansatz.
      symbols: list of `sympy.Symbol`s which are the parameters of the model.
    """
  symbols = []
  circuit = cirq.Circuit()
  for layer in range(num_layers):
    xz_circuit, xz_symbols = get_2d_xz_rotation_layer(rows, cols, layer, name)
    circuit += xz_circuit
    symbols += xz_symbols
    cnot_circuit, cnot_symbols = get_2d_cnot_exp_layer(rows, cols, layer, name)
    circuit += cnot_circuit
    symbols += cnot_symbols
  return circuit, symbols


# ============================================================================ #
# Trotter components.
# ============================================================================ #


def get_trotter_model_unitary(p, h_list, name):
  """Get a trotterized ansatz.

    Args:
      p: integer representing the number of QAOA steps.
      h_list: List of `cirq.PauliSum`s representing the Hamiltonians to
        exponentiate to build the circuit.
      name: string used to make symbols unique to this call.

    Returns:
      circuit: `cirq.Circuit` representing the parameterized QAOA ansatz.
      all_symbols: Python `list` of `sympy.Symbol`s containing all the
          parameters of the circuit.
    """
  circuit = cirq.Circuit()
  all_symbols = []
  for j in range(p):
    for n, h in enumerate(h_list):
      new_symb = sympy.Symbol("phi_{0}_L{1}_H{2}".format(name, j, n))
      circuit += tfq.util.exponential([h], coefficients=[new_symb])
      all_symbols.append(new_symb)
  return circuit, all_symbols


# ============================================================================ #
# QNHF components.
# ============================================================================ #


def get_general_trotter_unitary(symbol_array, h_list):
  """
    Args:
      symbol_array: 2-D array (list of lists) of `sympy.Symbol`s to use when
        exponentiating the Hamiltonians.  The first index is the trotter layer,
        the second index is corresponding Hamiltonian in `h_list`.
      h_list: List of `cirq.PauliSum`s representing all the Hamiltonians to
        exponentiate in a single trotter step.

    Returns:
      circuit: `cirq.Circuit` representing the parameterized trotter ansatz.
    """
  if len(symbol_array[0]) != len(h_list):
    raise ValueError(
        "Must have the same number of symbols as Hamiltonians in each layer.")
  circuit = cirq.Circuit()
  for s_layer in symbol_array:
    for n, h in enumerate(h_list):
      circuit += tfq.util.exponential([h], coefficients=[s_layer[n]])
  return circuit


def get_qnhf_symbols(p, n_bits, n_h, max_k, name):
  """Get all the symbols used by QNHF QHBMs.

    Args:
      p: the number of trotter steps.
      n_bits: the number of bits in the discrete sample space.
      max_k: the maximum locality of interactions in the classical model.
      n_h: the number of non-classical hamiltonian terms.
      name: string appended to symbols to uniqueify across QHBMs.

    Returns:
      eta_theta_symbols: 2-D array of `sympy.Symbol`s, where the first index is
      the
        trotter step and the second index is the particular symbol at that
        layer.
      phis_symbols: 2-D array of `sympy.Symbol`s, where the first index is the
        trotter step and the second index is the hamiltonian term at that layer.
    """
  eta_theta_symbols = []
  phis_symbols = []
  num_thetas_per_layer = ebm.get_klocal_energy_function_num_values(
      n_bits, max_k)
  for t_step in range(p):
    eta_theta_symbols.append([])
    for j in range(num_thetas_per_layer):
      eta_theta_symbols[-1].append(
          sympy.Symbol("eta_L{1}_theta_T{2}_{0}".format(name, t_step, j)))
    phis_symbols.append([])
    for m in range(n_h):
      phis_symbols[-1].append(
          sympy.Symbol("phi_L{1}_H{2}_{0}".format(name, t_step, m)))
  return eta_theta_symbols, phis_symbols


def get_qnhf_diagonal_operators(qubits, max_k):
  diag_op_list = []
  for k in range(1, max_k + 1):
    index_list = ebm.get_parity_index_list(len(qubits), k)
    for this_index_list in index_list:
      this_z_list = [cirq.Z(qubits[i]) for i in this_index_list]
      diag_op_list.append(
          cirq.PauliSum.from_pauli_strings(cirq.PauliString(*this_z_list)))
  return diag_op_list


def get_qnhf_model_unitary(p, qubits, max_k, h_list, name):
  """Get the QNHF unitary corresponding to the given Hamiltonians.

    Args:
      p: the number of trotter steps.
      qubits: list of `cirq.GridQubit`s on which to build KOBE.
      max_k: the maximum locality of interactions in the classical model.
      h_list: list of `cirq.PauliSum`s representing the hamiltonian terms.
      name: string appended to symbols to uniqueify across QHBMs.

    Returns:
      circuit: `cirq.Circuit` representing the QNHF ansatz.
    """
  eta_theta_symbols, phis_symbols = get_qnhf_symbols(p, len(qubits),
                                                     len(h_list), max_k, name)
  total_symbols = []
  for t_layer, p_layer in zip(eta_theta_symbols, phis_symbols):
    total_symbols.append(t_layer + p_layer)
  classical_h_list = get_qnhf_diagonal_operators(qubits, max_k)
  total_h_list = classical_h_list + h_list
  return get_general_trotter_unitary(total_symbols, total_h_list)


# ============================================================================ #
# Convolutional components.
# ============================================================================ #


def qubits_to_grid(qubits):
  qubit_grid = []
  for q in qubits:
    if q.row > len(qubit_grid) - 1:
      qubit_grid.append([])
    bisect.insort(qubit_grid[q.row], q)
  return qubit_grid


def one_qubit_unitary(q, symbols):
  """Make a Cirq circuit for an arbitrary one qubit unitary."""
  return cirq.Circuit(
      cirq.X(q)**symbols[0],
      cirq.Y(q)**symbols[1],
      cirq.Z(q)**symbols[2])


def two_qubit_unitary(q0, q1, symbols):
  """Make a Cirq circuit for an arbitrary two qubit unitary."""
  circuit = cirq.Circuit()
  circuit += one_qubit_unitary(q0, symbols[0:3])
  circuit += one_qubit_unitary(q1, symbols[3:6])
  circuit += [cirq.ZZ(q0, q1)**symbols[6]]
  circuit += [cirq.YY(q0, q1)**symbols[7]]
  circuit += [cirq.XX(q0, q1)**symbols[8]]
  circuit += one_qubit_unitary(q0, symbols[9:12])
  circuit += one_qubit_unitary(q1, symbols[12:])
  return circuit


def two_qubit_pool(source_qubit, sink_qubit, symbols):
  """Make a Cirq circuit to do a parameterized 'pooling' operation, which

    attempts to reduce entanglement down from two qubits to just one.
  """
  pool_circuit = cirq.Circuit()
  sink_basis_selector = one_qubit_unitary(sink_qubit, symbols[0:3])
  source_basis_selector = one_qubit_unitary(source_qubit, symbols[3:6])
  pool_circuit.append(sink_basis_selector)
  pool_circuit.append(source_basis_selector)
  pool_circuit.append(cirq.CNOT(control=source_qubit, target=sink_qubit))
  pool_circuit.append(sink_basis_selector**-1)
  return pool_circuit


# TODO(#19)
# def quantum_convolutional_layer(qubits, layer_num, name):
#     """Assumes the qubits are arranged on a grid."""
#     qubit_grid = qubits_to_grid(qubits)
#     layer_symbols = []
#     circuit = cirq.Circuit()
#     tied_2q_symbols = [
#         sympy.Symbol("s_conv_I{0}_L{1}_N{2}".format(name, layer_num, s_num))
#         for s_num in range(15)
#     ]
#     # Apply horizontal bonds
#     for r in qubit_grid:
#         r_clipped = r[1:]
#         for alt_r in [r, r_clipped]:
#             for q0, q1 in zip(alt_r[::2], alt_r[1::2]):
#                 circuit += two_qubit_unitary(q0, q1, tied_2q_symbols)
#     # Apply vertical bonds
#     grid_clipped = qubit_grid[1:]
#     for r0, r1 in zip(qubit_grid[::2], qubit_grid[1::2]):
#         for q0, q1 in zip(r0, r1):
#             circuit += two_qubit_unitary(q0, q1, tied_2q_symbols)
#     for r0, r1 in zip(grid_clipped[::2], grid_clipped[1::2]):
#         for q0, q1 in zip(r0, r1):
#             circuit += two_qubit_unitary(q0, q1, tied_2q_symbols)
#     return circuit, tied_2q_symbols

# def quantum_pool_layer(qubits, layer_num, name, direction_flag):
#     """Assumes the qubits are arranged on a grid."""
#     qubit_grid = qubits_to_grid(qubits)
#     layer_symbols = []
#     circuit = cirq.Circuit()
#     tied_pool_symbols = [
#         sympy.Symbol("s_pool_I{0}_L{1}_N{2}".format(name, layer_num, s_num))
#         for s_num in range(6)
#     ]



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for the energy_model module."""

import itertools
from typing import List

import tensorflow as tf


def check_bits(bits: List[int]) -> List[int]:
  """Confirms the input is a valid bit index list."""
  if len(set(bits)) != len(bits):
    raise ValueError("All entries of `bits` must be unique.")
  return bits


def check_order(order: int) -> int:
  """Confirms the input is a valid parity order."""
  if not isinstance(order, int):
    raise TypeError("`order` must be an integer.")
  if order <= 0:
    raise ValueError("`order` must be greater than zero.")
  return order


class SpinsFromBitstrings(tf.keras.layers.Layer):
  """Simple layer taking bits to spins."""

  def __init__(self):
    """Initializes a SpinsFromBitstrings."""
    super().__init__(trainable=False)

  def call(self, inputs):
    """Returns the spins corresponding to the input bitstrings.

    Note that this maps |0> -> +1 and |1> -> -1.  This is in accordance with
    the usual interpretation of the Bloch sphere.
    """
    return tf.cast(1 - 2 * inputs, tf.float32)


class VariableDot(tf.keras.layers.Layer):
  """Utility layer for dotting input with a same-sized variable."""

  def __init__(self,
               initializer: tf.keras.initializers.Initializer = tf.keras
               .initializers.RandomUniform()):
    """Initializes a VariableDot layer.

    Args:
      initializer: A `tf.keras.initializers.Initializer` which specifies how to
        initialize the values of the parameters.
    """
    super().__init__()
    self._initializer = initializer

  def build(self, input_shape):
    """Initializes the internal variables."""
    self.kernel = tf.Variable(
        initial_value=self._initializer((input_shape[-1],)),
        name="kernel",
        trainable=True)

  def call(self, inputs):
    """Returns the dot product between the inputs and this layer's variables."""
    return tf.reduce_sum(inputs * self.kernel, -1)


class Parity(tf.keras.layers.Layer):
  """Computes the parities of input spins."""

  def __init__(self, bits: List[int], order: int):
    """Initializes a Parity layer."""
    super().__init__(trainable=False)
    bits = check_bits(bits)
    order = check_order(order)
    indices_list = []
    for i in range(1, order + 1):
      combos = itertools.combinations(range(len(bits)), i)
      indices_list.extend(list(combos))
    self.indices = tf.ragged.stack(indices_list)
    self.num_terms = len(indices_list)

  def call(self, inputs):
    """Returns a batch of parities corresponding to the input bitstrings."""
    parities_t = tf.zeros([self.num_terms, tf.shape(inputs)[0]])
    for i in tf.range(self.num_terms):
      parity = tf.reduce_prod(tf.gather(inputs, self.indices[i], axis=-1), -1)
      parities_t = tf.tensor_scatter_nd_update(parities_t, [[i]], [parity])
    return tf.transpose(parities_t)



# pylint: skip-file
# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions."""

import numpy as np

import tensorflow as tf
import tensorflow_quantum as tfq

# ============================================================================ #
# Density matrix utilities.
# ============================================================================ #


def pure_state_tensor_to_density_matrix(pure_states, counts):
  """Returns the uniform mixture of the given tensor of pure states.

    This function first treats every state as a column vector, then takes the
    outer product of each state with its adjoint.  Then the sum over all the
    outer
    products is taken, and the resulting matrix normalized.

    Args:
      pure_states: 2-D `tf.Tensor` of dtype `complex64` and shape [num_states,
        2**num_qubits] containing a pure state decomposition of a density
        operator.
  """
  expanded_s = tf.expand_dims(pure_states, 1)
  col_s = tf.transpose(expanded_s, [0, 2, 1])
  adj_s = tf.linalg.adjoint(col_s)
  prod = tf.linalg.matmul(col_s, adj_s)
  thermal_state = tf.reduce_sum(
      prod *
      tf.cast(tf.expand_dims(tf.expand_dims(counts, 1), 2), tf.complex64), 0)
  return tf.math.divide(thermal_state,
                        tf.cast(tf.reduce_sum(counts), tf.complex64))


def circuits_and_counts_to_density_matrix(circuits, counts):
  pure_states = tfq.layers.State()(circuits).to_tensor()
  return pure_state_tensor_to_density_matrix(pure_states, counts)


def fidelity(rho, sigma):
  """Calculate the fidelity between the two given density matrices.

    Args:
      rho: 2-D `tf.Tensor` of dtype `complex64` representing the left density
        matrix in the fidelity calculation.
      sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
        matrix in the fidelity calculation.

    Returns:
      A tf.Tensor float64 fidelity scalar between the two given density
      matrices.
    """
  e_rho, v_rho = tf.linalg.eigh(tf.cast(rho, tf.complex128))
  rho_sqrt = tf.linalg.matmul(
      tf.linalg.matmul(v_rho, tf.linalg.diag(tf.math.sqrt(e_rho))),
      tf.linalg.adjoint(v_rho),
  )
  omega = tf.linalg.matmul(
      tf.linalg.matmul(rho_sqrt, tf.cast(sigma, tf.complex128)), rho_sqrt)
  e_omega = tf.linalg.eigvals(omega)
  return tf.math.abs(tf.math.reduce_sum(tf.math.sqrt(e_omega)))**2


def fast_fidelity(rho, sigma):
  """Calculate the 10x faster fidelity between the two given density matrices.

    Args:
      rho: 2-D `tf.Tensor` of dtype `complex64` representing the left density
        matrix in the fidelity calculation.
      sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
        matrix in the fidelity calculation.

    Returns:
      A tf.Tensor float64 fidelity scalar between the two given density
      matrices.
    """
  e_rho, v_rho = tf.linalg.eigh(tf.cast(rho, tf.complex128))
  # Optimization
  # 1) tf.matmul(a, tf.linalg.diag(b))
  # -> tf.multiply(a, tf.tile(tf.expand_dims(b, axis=0), [tf.shape(a)[0], 1]))
  # minor portion, but 10x
  sqrt_e_rho = tf.sqrt(e_rho)
  v_rho_sqrt_e_rho = tf.multiply(
      v_rho,
      tf.tile(tf.expand_dims(sqrt_e_rho, axis=0), (tf.shape(v_rho)[0], 1)))
  rho_sqrt = tf.linalg.matmul(v_rho_sqrt_e_rho, tf.linalg.adjoint(v_rho))
  omega = tf.linalg.matmul(
      tf.linalg.matmul(rho_sqrt, tf.cast(sigma, tf.complex128)), rho_sqrt)
  # 2) Optimization eigvals -> eigh (10x), with numerical errors in 1e-4.
  e_omega, _ = tf.linalg.eigh(omega)
  return tf.math.abs(tf.math.reduce_sum(tf.math.sqrt(e_omega)))**2


def _np_fidelity_internal(e_rho, v_rho, sigma):
  """Internal np fidelity logic for preventing retracing graphs."""
  sqrt_e_rho = tf.cast(tf.sqrt(e_rho), dtype=tf.complex128)
  v_rho_sqrt_e_rho = tf.cast(
      tf.multiply(
          v_rho,
          tf.tile(tf.expand_dims(sqrt_e_rho, axis=0), (tf.shape(v_rho)[0], 1))),
      tf.complex128,
  )
  rho_sqrt = tf.linalg.matmul(v_rho_sqrt_e_rho, tf.linalg.adjoint(v_rho))
  omega = tf.linalg.matmul(
      tf.linalg.matmul(rho_sqrt, tf.cast(sigma, tf.complex128)), rho_sqrt)
  return omega


def np_fidelity(rho, sigma):
  """Calculate the numpy-based fidelity between the two given density matrices.

    tf.linalg.eigh() currently shows memory usage growth,
    so it's slower & requires larger memory than np.linalg.eigh().
    To enable the experiments, we need this function.

    For example, 4x3 HEA VQT or 3x3 QAIA VQT can't run fidelity calculation on
    64 GiB due to memory exceeds.

    Args:
      rho: 2-D `tf.Tensor` of dtype `complex64` representing the left density
        matrix in the fidelity calculation.
      sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
        matrix in the fidelity calculation.

    Returns:
      A tf.Tensor float64 fidelity scalar between the two given density
      matrices.
    """
  e_rho, v_rho = np.linalg.eigh(tf.cast(rho, dtype=tf.complex128))
  omega = _np_fidelity_internal(e_rho, v_rho, sigma)
  e_omega, _ = np.linalg.eigh(omega)
  return tf.math.abs(tf.math.reduce_sum(tf.math.sqrt(e_omega)))**2


def optimized_trace_matmul(rho, sigma):
  """Returns optimized version of tf.linalg.trace(tf.matmul(rho, sigma)).

    Assuming the both have the same shape.

    Args:
      rho: 2-D `tf.Tensor` of dtype `complex64` representing the left density
        matrix in the trace-matmul calculation.
      sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
        matrix in the trace-matmul calculation.

    Returns:
      A tf.Tensor float64 trace value between the two given density matrices.
    """
  return tf.reduce_sum(
      tf.multiply(
          tf.cast(rho, tf.complex128),
          tf.transpose(tf.cast(sigma, tf.complex128))))


def relative_entropy(rho, sigma):
  """Calculate the relative entropy between the two given density matrices.

    D(rho||sigma) = Tr[rho(log(rho) - log(sigma))]
                  = tf.linalg.trace(
                        tf.matmul(rho,
                                  tf.linalg.logm(rho) - tf.linalg.logm(sigma)))

    Args:
      rho: 2-D `tf.Tensor` of dtype `complex64` representing the left density
        matrix in the fidelity calculation.
      sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right density
        matrix in the fidelity calculation.

    Returns:
      A tf.Tensor float64 fidelity scalar between the two given density
      matrices.
    """
  log_rho = tf.linalg.logm(tf.cast(rho, tf.complex128))
  log_sigma = tf.linalg.logm(tf.cast(sigma, tf.complex128))
  return optimized_trace_matmul(rho, tf.subtract(log_rho, log_sigma))


def get_thermal_state(beta, h_num):
  """Computes the thermal state.

    Compute exp(-beta*h_num)/Tr[exp(-beta*h_num)].  Uses the log-sum-exp trick
    to
    avoid numerical instability.  When x is a vector with elements
    x_1, x_2, ..., x_N, define the log-sum-exp function as
    LSE(x) = ln(sum_i exp(x_i)).  Then, consider the entries of the derivative
    of
    LSE, [d/dx LSE(x)]_i = exp(x_i)/(sum_j exp(x_j)).  When h_u is the unitary
    diagonalizing h_num and h_diag is the corresponding matrix of eigenvalues,
    the thermal state can be rewritten as
    exp(-beta*h_num)/Tr[exp(-beta*h_num)]
        = exp(-beta*h_u@h_diag@h_u_dag)/Tr[exp(-beta*h_u@h_diag@h_u_dag)]
        = h_u @ exp(-beta* h_diag) @ h_u_dag /Tr[exp(-beta*h_u@h_diag@h_u_dag)]
          # Due to matrix exponential identity
          # https://en.wikipedia.org/wiki/Matrix_exponential#Diagonalizable_case
        = h_u @ exp(-beta* h_diag) @ h_u_dag /Tr[exp(-beta*h_diag)]
          # Due to cyclicity of trace
        = h_u @ exp(-beta* h_diag) @ h_u_dag / (sum_i <i|exp(-beta*h_diag)|i>)
          # Definition of trace
        = h_u @ exp(-beta* h_diag) @ h_u_dag / (sum_i exp(-beta*h_diag[i, i]))
          # Definition of diagonal matrix
        = h_u @ (exp(-beta* h_diag) / sum_i exp(-beta*h_diag[i, i])) @ h_u_dag
          # Scalar
        = h_u @ sum_i (exp(x_i) / sum_j exp(x_j))|i><i| @ h_u_dag
          # Let x_i = - beta * h_diag[i, i]
        = h_u @ sum_i [d/dx LSE(x)]_i|i><i| @ h_u_dag
          # Substitute LSE derivative

    Args:
      beta: Scalar `tf.Tensor` of dtype `float64` which is the inverse
        temperature at which to calculate the thermal state.
      h_num: Square, Hermitian `tf.Tensor` of dtype `complex128` which is the
        Hamiltonian whose thermal state is to be calculated.

    Returns:
      `tf.Tensor` of dtype `complex128` which is the density matrix of the
      thermal
          state of `h_num` at inverse temperature `beta`.
    """
  e_raw, v_raw = tf.linalg.eigh(h_num)
  e = tf.cast(e_raw, tf.float64)
  x = -1.0 * beta * e
  with tf.GradientTape() as g:
    g.watch(x)
    lse = tf.reduce_logsumexp(x)
  lse_grad = g.gradient(lse, x)
  tiled_lse_grad = tf.cast(
      tf.tile(tf.expand_dims(lse_grad, 0), [tf.shape(h_num)[0], 1]),
      tf.complex128)
  return tf.linalg.matmul(
      tf.math.multiply(v_raw, tiled_lse_grad), tf.linalg.adjoint(v_raw))


def _np_get_thermal_state_internal(beta, e, h_shape):
  x = -1.0 * beta * e
  with tf.GradientTape() as g:
    g.watch(x)
    lse = tf.reduce_logsumexp(x)
  lse_grad = g.gradient(lse, x)
  tiled_lse_grad = tf.cast(
      tf.tile(tf.expand_dims(lse_grad, 0), [h_shape[0], 1]), tf.complex128)
  return tiled_lse_grad


def np_get_thermal_state(beta, h_num):
  """Computes the thermal state with numpy eigh().

    For details, please refer to get_thermal_state().
    Since tf.linalg.eigh() shows huge memory usage growth, we can't use it
    for larger grid size.

    Args:
      beta: Scalar `tf.Tensor` of dtype `float64` which is the inverse
        temperature at which to calculate the thermal state.
      h_num: Square, Hermitian `tf.Tensor` of dtype `complex128` which is the
        Hamiltonian whose thermal state is to be calculated.

    Returns:
      `tf.Tensor` of dtype `complex128` which is the density matrix of the
      thermal
          state of `h_num` at inverse temperature `beta`.
    """
  h_num = tf.cast(h_num, tf.complex128)
  e_raw, v_raw = np.linalg.eigh(h_num)
  tiled_lse_grad = _np_get_thermal_state_internal(
      tf.cast(beta, tf.float64), tf.cast(e_raw, tf.float64), tf.shape(h_num))
  return tf.linalg.matmul(
      tf.math.multiply(v_raw, tiled_lse_grad), tf.linalg.adjoint(v_raw))


def log_partition_function(beta, h_num):
  """Computes the logarithm of the partition function.

    The log partition function is ln(Tr[expm(-beta * h_num)])

    Args:
      beta: Scalar `tf.Tensor` of dtype `float64` which is the inverse
        temperature at which to calculate the thermal state.
      h_num: Square, Hermitian `tf.Tensor` of dtype `complex128` which is the
        Hamiltonian appearing in the partition function.

    Returns:
      Scalar `tf.Tensor` of dtype `float64` which is the logarithm of the
        partition function.
    """
  # Use eigh instead of eigvalsh to ease backpropagation, see docs on eigvalsh.
  h_eigs, _ = np.linalg.eigh(tf.cast(
      h_num, tf.complex128))  # h_eigs are real since h_num is Hermitian
  return tf.reduce_logsumexp(-tf.cast(beta, tf.float64) *
                             tf.cast(h_eigs, tf.float64))


def entropy(rho):
  """Computes the von Neumann entropy of the given density matrix.

    The von Neumann entropy is -Tr[rho ln[rho]].  Note that the entropy is then
    in
    units of nats here instead of bits since we use natural logarithm.
    Simplify:
    -Tr[rho ln[rho]]
        = -Tr[U D U_dag ln[U D U_dag]]
        = -Tr[U D U_dag U ln[D] U_dag]
        = -Tr[D ln[D]]
        = -sum_i D_ii ln D_ii

    Args:
      rho: Square, Hermitian `tf.Tensor` of dtype `complex128` which is the
        density matrix.

    Returns:
      Scalar `tf.Tensor` of dtype `float64` which is the von Neumann entropy (in
        units of nats) of `rho`.
    """
  # Use eigh instead of eigvalsh to ease backpropagation, see docs on eigvalsh.
  rho_eigs, _ = tf.linalg.eigh(rho)
  # limit as x->0 of x ln x is 0, so replace nans from any 0 ln 0 with 0
  rho_prod = tf.math.multiply_no_nan(tf.math.log(rho_eigs), rho_eigs)
  return -tf.math.reduce_sum(rho_prod)


def density_matrix_to_image(dm):
  """Convert multi-qubit density matrix into an RGB image."""
  max_qubits = 9
  total_edge = 2**max_qubits
  dm_len = tf.shape(dm)[0]
  superpixel = tf.linalg.LinearOperatorFullMatrix(
      tf.ones((total_edge // dm_len, total_edge // dm_len), dtype=tf.float32))
  my_zeros = tf.zeros([total_edge, total_edge, 1])
  dm_real = tf.math.abs(tf.math.real(dm))
  dm_imag = tf.math.abs(tf.math.imag(dm))
  max_dm_real = tf.math.reduce_max(dm_real)
  max_dm_imag = tf.math.reduce_max(dm_imag)
  max_dm = tf.math.maximum(max_dm_real, max_dm_imag)
  dm_real = tf.cast(dm_real / max_dm, tf.float32)
  dm_imag = tf.cast(dm_imag / max_dm, tf.float32)
  dm_real = tf.linalg.LinearOperatorKronecker(
      [tf.linalg.LinearOperatorFullMatrix(dm_real), superpixel]).to_dense()
  dm_imag = tf.linalg.LinearOperatorKronecker(
      [tf.linalg.LinearOperatorFullMatrix(dm_imag), superpixel]).to_dense()
  dm_real = tf.expand_dims(dm_real, 2)
  dm_imag = tf.expand_dims(dm_imag, 2)
  return tf.reshape(
      tf.concat([dm_real, my_zeros, dm_imag], 2),
      (1, total_edge, total_edge, 3))


# ============================================================================ #
# Bitstring utilities.
# ============================================================================ #


def qubits_to_indices(qubits):
  """Maps list of cirq.GridQubit to a tensor of the column and row coordinates.

  Args:
    qubits: Python `list` of `cirq.GridQubit`s.
  
  Returns:
    indices: `tf.Tensor` of shape [len(qubits), 2] where `indices[i][0]` and
      `indices[i][1]` give the row of and column of `qubits[i]`, respectively.
  """
  indices = tf.constant([[q.row, q.col] for q in qubits], dtype=tf.int32)
  return indices


def qubit_sub_indices(qubits_total, qubits_sublist):
  """Returns the indices of the qubit sublist in the qubit total.
  
  Args:
    qubits_total: `tf.Tensor` of shape [n, 2], the output of `qubits_to_indices`
      on some list of `cirq.GridQubit`s.
    qubits_sublist: `tf.Tensor` of shape [m, 2], the output of
      `qubits_to_indices` on some list of `cirq.GridQubit`s.

  Returns:
    sub_indices: `tf.Tensor` of shape [j], whose entries are the indices of the
      sublist of qubits in total list of qubits.  When `qubits_sublist` is a
      subset of `qubits_total`, then j == m.  More generally, j is the number of
      qubits in `qubits_sublist` that also appear in `qubits_total`.
  """
  qubits_total_tiled = tf.tile(
      tf.expand_dims(qubits_total, 0), [tf.shape(qubits_sublist)[0], 1, 1])
  qubits_sublist_tiled = tf.tile(
      tf.expand_dims(qubits_sublist, 1), [1, tf.shape(qubits_total)[0], 1])
  qubit_overlaps = tf.equal(qubits_total_tiled, qubits_sublist_tiled)
  reduced_qubit_overlaps = tf.reduce_all(qubit_overlaps, 2)
  doubled_indices = tf.where(reduced_qubit_overlaps)
  sub_indices = tf.transpose(doubled_indices)[1]
  return sub_indices


@tf.function
def unique_bitstrings_with_counts(bitstrings):
  """Extract the unique bitstrings in the given bitstring tensor.

    Works by converting each bitstring to a 64 bit integer, then using built-in
    `tf.unique_with_counts` on this 1-D array, then mapping these integers back
    to bitstrings. The inputs and outputs are to be related by the same
    invariants as those of `tf.unique_with_counts`,
    y[idx[i]] = input_bitstrings[i] for i in [0, ..., rank(input_bitstrings)-1]

    TODO(zaqqwerty): the signature and return values are designed to be similar
    to those of tf.unique_with_counts.  This function is needed because
    `tf.unique_with_counts` does not work on 2-D tensors.  When it begins to
    work on 2-D tensors, then this function will be deprecated.

    Args:
      input_bitstrings: 2-D `tf.Tensor` of dtype `int8`.  This tensor is
        interpreted as a list of bitstrings.  Bitstrings are required to be 64
        bits or fewer.
      out_idx: An optional `tf.DType` from: `tf.int32`, `tf.int64`. Defaults to
        `tf.int32`.  Specified type of idx and count outputs.

    Returns:
      y: 2-D `tf.Tensor` of dtype `int8` containing the unique 0-axis entries of
        `input_bitstrings`.
      count: 1-D `tf.Tensor` of dtype `out_idx` such that `count[i]` is the
        number of occurences of `y[i]` in `input_bitstrings`.
  """
  # Convert bitstrings to integers and uniquify those integers.
  input_shape = tf.shape(bitstrings)
  mask = tf.cast(bitstrings, tf.int64)
  base = tf.bitwise.left_shift(
      mask, tf.range(tf.cast(input_shape[1], tf.int64), dtype=tf.int64))
  ints_equiv = tf.reduce_sum(base, 1)
  _, idx, counts = tf.unique_with_counts(ints_equiv)

  # Convert unique integers to corresponding unique bitstrings.
  unique_bitstrings = tf.zeros((tf.shape(counts)[0], input_shape[1]),
                               dtype=tf.int8)
  unique_bitstrings = tf.tensor_scatter_nd_update(unique_bitstrings,
                                                  tf.expand_dims(idx, axis=1),
                                                  bitstrings)

  return unique_bitstrings, counts



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for inference on energy functions represented by a BitstringEnergy."""

import abc
import itertools
from typing import Union

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability import distributions as tfd

from qhbmlib import energy_model


class EnergyInference(tf.keras.layers.Layer, abc.ABC):
  """Sets the methods required for inference on BitstringEnergy objects."""

  def __init__(self, name: Union[None, str] = None):
    """Initializes an EnergyInference.

    Args:
      name: Optional name for the model.
    """
    super().__init__(name=name)

  @abc.abstractmethod
  def infer(self, energy: energy_model.BitstringEnergy):
    """Do the work to ready this layer for use.

    This should be called each time the underlying model is updated.

    Args:
      energy: The parameterized energy function which defines this distribution
        via the equations of an energy based model.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def sample(self, n):
    """Returns samples from the EBM corresponding to `self.energy`.

    This can be an approximate sampling.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def entropy(self):
    """Returns an estimate of the entropy."""
    raise NotImplementedError()

  @abc.abstractmethod
  def log_partition(self):
    """Returns an estimate of the log partition function."""
    raise NotImplementedError()

  def call(self, inputs):
    """Returns the number of samples specified in the inputs."""
    return self.sample(inputs)


class AnalyticEnergyInference(EnergyInference):
  """Uses an explicit categorical distribution to implement parent functions."""

  def __init__(self, num_bits: int, name: Union[None, str] = None):
    """Initializes an AnalyticEnergyInference.

    Internally, this class saves all possible bitstrings as a tensor, whose
    energies are calculated relative to an input energy function for sampling
    and other inference tasks.

    Args:
      num_bits: Number of bits on which this layer acts.
      name: Optional name for the model.
    """
    super().__init__(name=name)
    self._all_bitstrings = tf.constant(
        list(itertools.product([0, 1], repeat=num_bits)), dtype=tf.int8)
    self._dist_realization = tfp.layers.DistributionLambda(
        make_distribution_fn=lambda t: tfd.Categorical(logits=-1 * t))
    self._current_dist = None

  @property
  def all_bitstrings(self):
    """Returns every bitstring."""
    return self._all_bitstrings

  @property
  def all_energies(self):
    """Returns the energy of every bitstring."""
    return self.energy(self.all_bitstrings)

  @property
  def current_dist(self):
    """Bernoulli distribution set during last call to `self.infer`."""
    return self._current_dist

  def infer(self, energy: energy_model.BitstringEnergy):
    """See base class docstring."""
    self.energy = energy
    x = tf.squeeze(self.all_energies)
    self._current_dist = self._dist_realization(x)

  def sample(self, n):
    """See base class docstring"""
    return tf.gather(self.all_bitstrings, self._current_dist.sample(n), axis=0)

  def entropy(self):
    """See base class docstring"""
    return self._current_dist.entropy()

  def log_partition(self):
    """See base class docstring"""
    # TODO(#115)
    return tf.reduce_logsumexp(self._current_dist.logits_parameter())

  def call(self, inputs):
    if self._current_dist is None:
      raise RuntimeError("`infer` must be called at least once.")
    if inputs is None:
      return self._current_dist
    else:
      return self.sample(inputs)


class BernoulliEnergyInference(EnergyInference):
  """Manages inference for a Bernoulli defined by spin energies."""

  def __init__(self, name: Union[None, str] = None):
    """Initializes a BernoulliEnergyInference.

    Args:
      name: Optional name for the model.
    """
    super().__init__(name=name)
    self._dist_realization = tfp.layers.DistributionLambda(
        make_distribution_fn=lambda t: tfd.Bernoulli(logits=t, dtype=tf.int8))
    self._current_dist = None

  @property
  def current_dist(self):
    """Categorical distribution set during last call to `self.infer`."""
    return self._current_dist

  def infer(self, energy: energy_model.BitstringEnergy):
    """See base class docstring."""
    self.energy = energy
    self._current_dist = self._dist_realization(self.energy.logits)

  def sample(self, n):
    """See base class docstring"""
    return self._current_dist.sample(n)

  def entropy(self):
    """Returns the exact entropy.

    The total entropy of a set of spins is the sum of each individual spin's
    entropies.
    """
    return tf.reduce_sum(self._current_dist.entropy())

  def log_partition(self):
    r"""Returns the exact log partition function.

    For a single spin of energy $\theta$, the partition function is
    $$Z_\theta = \exp(\theta) + \exp(-\theta).$$
    Since each spin is independent, the total log partition function is
    the sum of the individual spin log partition functions.
    """
    thetas = 0.5 * self.energy.logits
    single_log_partitions = tf.math.log(
        tf.math.exp(thetas) + tf.math.exp(-1.0 * thetas))
    return tf.math.reduce_sum(single_log_partitions)

  def call(self, inputs):
    if self._current_dist is None:
      raise RuntimeError("`infer` must be called at least once.")
    if inputs is None:
      return self._current_dist
    else:
      return self.sample(inputs)



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for the circuit_model module."""

from typing import List

import cirq
import sympy


def bit_circuit(qubits: List[cirq.GridQubit], name="bit_circuit"):
  """Returns exponentiated X gate on each qubit and the exponent symbols."""
  circuit = cirq.Circuit()
  for n, q in enumerate(qubits):
    bit = sympy.Symbol("{0}_bit_{1}".format(name, n))
    circuit += cirq.X(q)**bit
  return circuit



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for inference on quantum circuits represented by QuantumCircuit."""

from typing import Union

import cirq
import tensorflow as tf
import tensorflow_quantum as tfq

from qhbmlib import circuit_model
from qhbmlib import utils


class QuantumInference(tf.keras.layers.Layer):
  """Methods for inference on QuantumCircuit objects."""

  def __init__(self,
               backend: Union[str, cirq.Sampler] = "noiseless",
               differentiator: Union[None,
                                     tfq.differentiators.Differentiator] = None,
               name: Union[None, str] = None):
    """Initialize a QuantumInference layer.

    Args:
      backend: Specifies what backend TFQ will use to compute expectation
        values. `str` options are {"noisy", "noiseless"}; users may also specify
        a preconfigured cirq execution object to use instead.
      differentiator: Specifies how to take the derivative of a quantum circuit.
      name: Identifier for this inference engine.
    """
    self._differentiator = differentiator
    self._backend = backend
    self._sample_layer = tfq.layers.Sample(backend=backend)
    if backend == "noiseless":
      self._expectation_layer = tfq.layers.Expectation(
          backend=backend, differentiator=differentiator)

      def _expectation_function(circuits, symbol_names, symbol_values,
                                operators, *args):
        del args
        return self._expectation_layer(
            circuits,
            symbol_names=symbol_names,
            symbol_values=symbol_values,
            operators=operators)
    else:
      self._expectation_layer = tfq.layers.SampledExpectation(
          backend=backend, differentiator=differentiator)

      def _expectation_function(circuits, symbol_names, symbol_values,
                                operators, repetitions):
        return self._expectation_layer(
            circuits,
            symbol_names=symbol_names,
            symbol_values=symbol_values,
            operators=operators,
            repetitions=repetitions,
        )

    self._expectation_function = _expectation_function
    super().__init__(name=name)

  @property
  def backend(self):
    return self._backend

  @property
  def differentiator(self):
    return self._differentiator

  def expectation(self,
                  qnn: circuit_model.QuantumCircuit,
                  initial_states: tf.Tensor,
                  counts: tf.Tensor,
                  operators: tf.Tensor,
                  reduce: bool = True):
    """Returns the expectation values of the operators against the QNN.

      Args:
        qnn: The parameterized quantum circuit on which to do inference.
        initial_states: Shape [batch_size, num_qubits] of dtype `tf.int8`.
          These are the initial states of each qubit in the circuit.
        counts: Shape [batch_size] of dtype `tf.int32` such that `counts[i]` is
          the weight of `initial_states[i]` when computing expectations.
          Additionally, if `self.backend != "noiseless", `counts[i]` samples
          are drawn from `(qnn)|initial_states[i]>` and used to compute
          the the corresponding expectation.
        operators: `tf.Tensor` of strings with shape [n_ops], result of calling
          `tfq.convert_to_tensor` on a list of cirq.PauliSum, `[op1, op2, ...]`.
          Will be tiled to measure `<op_j>_((qnn)|initial_states[i]>)`
          for each i and j.
        reduce: bool flag for whether or not to average over i.

      Returns:
        If `reduce` is true, a `tf.Tensor` with shape [n_ops] whose entries are
        are the batch-averaged expectation values of `operators`.
        Else, a `tf.Tensor` with shape [batch_size, n_ops] whose entries are the
        unaveraged expectation values of each `operator` against each `circuit`.
      """
    circuits = qnn(initial_states)
    num_circuits = tf.shape(circuits)[0]
    num_operators = tf.shape(operators)[0]
    tiled_values = tf.tile(
        tf.expand_dims(qnn.symbol_values, 0), [num_circuits, 1])
    tiled_operators = tf.tile(tf.expand_dims(operators, 0), [num_circuits, 1])
    expectations = self._expectation_function(
        circuits,
        qnn.symbol_names,
        tiled_values,
        tiled_operators,
        tf.tile(tf.expand_dims(counts, 1), [1, num_operators]),
    )
    if reduce:
      return utils.weighted_average(counts, expectations)
    return expectations

  def sample(self, qnn: circuit_model.QuantumCircuit, initial_states: tf.Tensor,
             counts: tf.Tensor):
    """Returns bitstring samples from the QNN.

      Args:
        qnn: The parameterized quantum circuit on which to do inference.
        initial_states: Shape [batch_size, num_qubits] of dtype `tf.int8`.
          These are the initial states of each qubit in the circuit.
        counts: Shape [batch_size] of dtype `tf.int32` such that `counts[i]` is
          the number of samples to draw from `(qnn)|initial_states[i]>`.

      Returns:
        ragged_samples: `tf.RaggedTensor` of DType `tf.int8` structured such
          that `ragged_samples[i]` contains `counts[i]` bitstrings drawn from
          `(qnn)|initial_states[i]>`.
    """
    circuits = qnn(initial_states)
    num_circuits = tf.shape(circuits)[0]
    tiled_values = tf.tile(
        tf.expand_dims(qnn.symbol_values, 0), [num_circuits, 1])
    num_samples_mask = tf.cast((tf.ragged.range(counts) + 1).to_tensor(),
                               tf.bool)
    num_samples_mask = tf.map_fn(tf.random.shuffle, num_samples_mask)
    samples = self._sample_layer(
        circuits,
        symbol_names=qnn.symbol_names,
        symbol_values=tiled_values,
        repetitions=tf.expand_dims(tf.math.reduce_max(counts), 0))
    return tf.ragged.boolean_mask(samples, num_samples_mask)



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities used across more than one file."""

import tensorflow as tf


class Squeeze(tf.keras.layers.Layer):
  """Wraps tf.squeeze in a Keras Layer."""

  def __init__(self, axis=None):
    """Initializes a Squeeze layer.

    Args:
      axis: An optional list of ints. Defaults to []. If specified, only
        squeezes the dimensions listed. The dimension index starts at 0. It is
        an error to squeeze a dimension that is not 1. Must be in the range
        [-rank(input), rank(input)). Must be specified if input is
        a RaggedTensor.
    """
    super().__init__()
    if axis is None:
      axis = []
    self._axis = axis

  def call(self, inputs):
    """Applies tf.squeeze to the inputs."""
    return tf.squeeze(inputs, axis=self._axis)


def weighted_average(counts: tf.Tensor, values: tf.Tensor):
  """Returns the weighted average of input values.

  Row `i` of `values` is multiplied by `counts[i]`, resulting in a weighted
  version of values; the mean is then taken across the first dimension.

  Args:
    counts: Non-negative integers of shape [batch_size].
    values: Floats of shape [batch_size, n].
  """
  expanded_counts = tf.expand_dims(tf.cast(counts, tf.float32), -1)
  weighted_values = expanded_counts * values
  return tf.reduce_sum(weighted_values, 0) / tf.reduce_sum(expanded_counts)



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for defining quantum Hamiltonian-based models."""

from typing import Union

import tensorflow as tf
import tensorflow_quantum as tfq

from qhbmlib import circuit_model
from qhbmlib import energy_model


class Hamiltonian(tf.keras.layers.Layer):
  """Diagonalized (spectral) representation of a Hermitian operator."""

  def __init__(self,
               energy: energy_model.BitstringEnergy,
               circuit: circuit_model.QuantumCircuit,
               name: Union[None, str] = None):
    """Initializes a Hamiltonian.

    Args:
      energy: Represents the eigenvalues of this operator.
      circuit: Represents the eigenvectors of this operator.
      name: Optional name for the model.
    """
    super().__init__(name=name)
    if energy.num_bits != len(circuit.qubits):
      raise ValueError(
          "`energy` and `circuit` must act on the same number of bits.")
    self.energy = energy
    self.circuit = circuit
    self.circuit_dagger = circuit**-1

    self.operator_shards = None
    if isinstance(self.energy, energy_model.PauliMixin):
      self.operator_shards = tfq.convert_to_tensor(
          self.energy.operator_shards(self.circuit.qubits))



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Hamiltonians to use with QHBMs."""

import random

import cirq


def _qubit_grid(rows, cols):
  """Rectangle of qubits returned as a nested list."""
  qubits = []
  for r in range(rows):
    qubits.append([cirq.GridQubit(r, c) for c in range(cols)])
  return qubits


def heisenberg_bond(q0, q1, bond_type):
  """Given two Cirq qubits, return the PauliSum that bonds them."""
  if bond_type == 1:
    bond = cirq.X
  elif bond_type == 2:
    bond = cirq.Y
  elif bond_type == 3:
    bond = cirq.Z
  else:
    raise ValueError("Unknown bond type.")
  return cirq.PauliSum.from_pauli_strings(
      [cirq.PauliString(bond(q0), bond(q1))])


def heisenberg_hamiltonian_shard(rows, columns, jh, jv, bond_type):
  """Returns a commuting subset of the 2D Heisenberg Hamiltonian."""
  qubits = _qubit_grid(rows, columns)
  heisenberg = cirq.PauliSum()
  # Apply horizontal bonds
  for r in qubits:
    for q0, q1 in zip(r, r[1::]):
      heisenberg += jh * heisenberg_bond(q0, q1, bond_type)
  # Apply vertical bonds
  for r0, r1 in zip(qubits, qubits[1::]):
    for q0, q1 in zip(r0, r1):
      heisenberg += jv * heisenberg_bond(q0, q1, bond_type)
  return heisenberg


def tfim_x_shard(rows, columns):
  """Build the X component of a rectangular toroid TFIM at critical field."""
  lambda_crit = 3.05  # see https://arxiv.org/abs/cond-mat/0703788
  qubits = cirq.GridQubit.rect(rows, columns)
  x_sum = cirq.PauliSum()
  for q in qubits:
    x_sum += lambda_crit * cirq.X(q)
  return x_sum


def tfim_zz_shard(rows, columns):
  """Uniform ferromagnetic interaction on a rectangular toroid."""
  qubits = _qubit_grid(rows, columns)
  extended_qubits = _qubit_grid(rows, columns)
  for r, row in enumerate(qubits):
    extended_qubits[r].append(row[0])
  extended_qubits.append(qubits[0])
  zz_sum = cirq.PauliSum()
  # Horizontal interactions.
  for row in extended_qubits[:-1]:
    for q0, q1 in zip(row, row[1:]):
      zz_sum += cirq.Z(q0) * cirq.Z(q1)
  # Vertical interactions.
  for row_0, row_1 in zip(extended_qubits, extended_qubits[1:]):
    for q0, q1 in zip(row_0[:-1], row_1):
      zz_sum += cirq.Z(q0) * cirq.Z(q1)
  return zz_sum


def tfim_x_shard_random(rows, columns, mean, std):
  """Build the X component of a rectangular toroid TFIM with random fields."""
  qubits = cirq.GridQubit.rect(rows, columns)
  x_sum = cirq.PauliSum()
  for q in qubits:
    x_sum += random.normalvariate(mean, std) * cirq.X(q)
  return x_sum


def tfim_zz_shard_random(rows, columns, mean, std):
  """Random ferromagnetic interaction on a rectangular toroid."""
  qubits = _qubit_grid(rows, columns)
  extended_qubits = _qubit_grid(rows, columns)
  for r, row in enumerate(qubits):
    extended_qubits[r].append(row[0])
  extended_qubits.append(qubits[0])
  zz_sum = cirq.PauliSum()
  # Horizontal interactions.
  for row in extended_qubits[:-1]:
    for q0, q1 in zip(row, row[1:]):
      zz_sum += random.normalvariate(mean, std) * cirq.Z(q0) * cirq.Z(q1)
  # Vertical interactions.
  for row_0, row_1 in zip(extended_qubits, extended_qubits[1:]):
    for q0, q1 in zip(row_0[:-1], row_1):
      zz_sum += random.normalvariate(mean, std) * cirq.Z(q0) * cirq.Z(q1)
  return zz_sum



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for inference on quantum Hamiltonians."""

from typing import Union

import tensorflow as tf

from qhbmlib import circuit_infer
from qhbmlib import energy_infer
from qhbmlib import hamiltonian_model
from qhbmlib import util


class QHBM(tf.keras.layers.Layer):
  """Methods for inference on Hamiltonian objects."""

  def __init__(self,
               e_inference: energy_infer.EnergyInference,
               q_inference: circuit_infer.QuantumInference,
               name: Union[None, str] = None):
    """Initializes a QHBM.

    Args:
      e_inference: Attends to density operator eigenvalues.
      q_inference: Attends to density operator eigenvectors.
      name: Optional name for the model.
    """
    super().__init__(name=name)
    self._e_inference = e_inference
    self._q_inference = q_inference

  @property
  def e_inference(self):
    """The object used for inference on density operator eigenvalues."""
    return self._e_inference

  @property
  def q_inference(self):
    """The object used for inference on density operator eigenvectors."""
    return self._q_inference

  def circuits(self, model: hamiltonian_model.Hamiltonian, num_samples: int):
    r"""Draws thermally distributed eigenstates from the model Hamiltonian.

    We define some terms that will be used to explain the algorithm.
    `model` is some modular Hamiltonian
    $$K_{\theta\phi} = U_\phi K_\theta U_\phi^\dagger.$$
    The [thermal state][1] corresponding to the model is
    $$ \rho_T = Z^{-1} e^{-\beta K_{\theta\phi}}.$$
    For QHBMs, we fix $\beta = 1$, effectively absorbing it into the definition
    of the modular Hamiltonian.  Then $\rho_T$ can be expanded as
    $$\rho_T = \sum_x p_\theta(x)U_\phi\ket{x}\bra{x}U_\phi^\dagger,$$
    where the probability is given by
    $$p_\theta(x) = \tr[\exp(-K_\theta)]\bra{x}\exp(-K_\theta)\ket{x}$$
    for $x\in\{1, \ldots, \dim(K_{\theta\phi})\} = \mathcal{X}$. Note that each
    $U_\phi\ket{x}$ is an eigenvector of both $\rho_T$ and $K_{\theta\phi}$.
    Corresponding to this operator is the [ensemble of quantum states][2]
    $$\mathcal{E} = \{p_\theta(x), U_\phi\ket{x}\}_{x\in\mathcal{X}}.$$
    This function returns pure state samples from the ensemble.

    We now explain the algorithm.  First, construct $X$ to be a classical
    random variable with probability distribution $p_\theta(x)$ set by
    `model.energy`.  Then, draw $n = $`num\_samples` bitstrings,
    $S=\{x_1, \ldots, x_n\}$, from $X$.  For each unique $x_i\in S$, set
    `states[i]` to the TFQ string representation of $U_\phi\ket{x_i}$, where
    $U_\phi$ is set by `model.circuit`.  Finally, set `counts[i]` equal to the
    number of times $x_i$ occurs in $S$.

    #### References
    [1]: Nielsen, Michael A. and Chuang, Isaac L. (2010).
         Quantum Computation and Quantum Information.
         Cambridge University Press.
    [2]: Wilde, Mark M. (2017).
         Quantum Information Theory (second edition).
         Cambridge University Press.

    Args:
      model: The modular Hamiltonian whose normalized exponential is the
        density operator governing the ensemble of states from which to sample.
      num_samples: Number of states to draw from the ensemble.

    Returns:
      states: 1D `tf.Tensor` of dtype `tf.string`.  Each entry is a TFQ string
        representation of an eigenstate of the Hamiltonian `model`.
      counts: 1D `tf.Tensor` of dtype `tf.int32`.  `counts[i]` is the number of
        times `states[i]` was drawn from the ensemble.
    """
    self.e_inference.infer(model.energy)
    samples = self.e_inference.sample(num_samples)
    bitstrings, counts = util.unique_bitstrings_with_counts(samples)
    states = model.circuit(bitstrings)
    return states, counts



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Impementations of the QMHL loss and its derivatives."""

import tensorflow as tf


def qmhl(qhbm_model, target_circuits, target_counts):
  """Calculate the QMHL loss of the qhbm model against the target.

  This loss is differentiable with respect to the trainable variables of the
    model.

  Args:
    qhbm_model: Parameterized model density operator.
    target_circuits: 1-D tensor of strings which are serialized circuits.
      These circuits represent samples from the data density matrix.
    target_counts: 1-D tensor of integers which are the number of samples to
      draw from the data density matrix: `target_counts[i]` is the number of
        samples to draw from `target_circuits[i]`.

  Returns:
    loss: Quantum cross entropy between the target and model.
  """

  @tf.custom_gradient
  def loss(trainable_variables):
    # We use `qhbm_model.ebm.trainable_variables` and
    # `qhbm_model.qnn.trainable_variables` instead
    del trainable_variables

    # log_partition estimate

    if qhbm_model.ebm.is_analytic:
      log_partition_function = qhbm_model.log_partition_function()
    else:
      bitstrings, _ = qhbm_model.ebm.sample(tf.reduce_sum(target_counts))
      energies = qhbm_model.ebm.energy(bitstrings)
      log_partition_function = tf.math.reduce_logsumexp(-1 * energies)

    # pulled back expectation of energy operator
    qnn_bitstrings, qnn_counts = qhbm_model.qnn.pulled_back_sample(
        target_circuits, target_counts)
    qnn_probs = tf.cast(qnn_counts, tf.float32) / tf.cast(
        tf.reduce_sum(qnn_counts), tf.float32)
    energies = qhbm_model.ebm.energy(qnn_bitstrings)
    avg_energy = tf.reduce_sum(qnn_probs * energies)

    def grad(grad_y, variables=None):
      """Gradients are computed using estimators from the QHBM paper."""
      # Thetas derivative.
      ebm_bitstrings, ebm_counts = qhbm_model.ebm.sample(
          tf.reduce_sum(target_counts))
      ebm_probs = tf.cast(ebm_counts, tf.float32) / tf.cast(
          tf.reduce_sum(ebm_counts), tf.float32)
      with tf.GradientTape() as tape:
        tape.watch(qhbm_model.ebm.trainable_variables)
        qnn_energies = qhbm_model.ebm.energy(qnn_bitstrings)
      # jacobian is a list over thetas, with ith entry a tensor of shape
      # [tf.shape(qnn_energies)[0], tf.shape(thetas[i])[0]]
      qnn_energy_jac = tape.jacobian(qnn_energies,
                                     qhbm_model.ebm.trainable_variables)

      with tf.GradientTape() as tape:
        tape.watch(qhbm_model.ebm.trainable_variables)
        ebm_energies = qhbm_model.ebm.energy(ebm_bitstrings)
      ebm_energy_jac = tape.jacobian(ebm_energies,
                                     qhbm_model.ebm.trainable_variables)

      # contract over bitstring weights
      grad_ebm = [
          grad_y *
          (tf.reduce_sum(
              tf.transpose(qnn_probs * tf.transpose(qnn_energy_grad)), 0) -
           tf.reduce_sum(
               tf.transpose(ebm_probs * tf.transpose(ebm_energy_grad)), 0))
          for qnn_energy_grad, ebm_energy_grad in zip(qnn_energy_jac,
                                                      ebm_energy_jac)
      ]

      # Phis derivative.
      if qhbm_model.ebm.has_operator:
        with tf.GradientTape() as tape:
          tape.watch(qhbm_model.qnn.trainable_variables)
          energy_shards = qhbm_model.qnn.pulled_back_expectation(
              target_circuits, target_counts, qhbm_model.operator_shards)
          energy = qhbm_model.ebm.operator_expectation(energy_shards)
        grad_qnn = tape.gradient(energy, qhbm_model.qnn.trainable_variables)
        grad_qnn = [grad_y * g for g in grad_qnn]
      else:
        raise NotImplementedError(
            "Derivative when EBM has no operator is not yet supported.")
      grad_qhbm = grad_ebm + grad_qnn
      if variables is None:
        return grad_qhbm
      return grad_qhbm, [tf.zeros_like(g) for g in grad_qhbm]

    return avg_energy + log_partition_function, grad

  return loss(qhbm_model.trainable_variables)



# pylint: skip-file
# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Module for defining and sampling from orthogonal sets of QNNs."""
import numbers
from typing import Any, Callable, Iterable, List, Union

import cirq
import sympy
import numpy as np
import tensorflow as tf
import tensorflow_quantum as tfq
from qhbmlib import util


def bit_circuit(qubits, name="bit_circuit"):
  """Returns exponentiated X gate on each qubit and the exponent symbols."""
  circuit = cirq.Circuit()
  for n, q in enumerate(qubits):
    bit = sympy.Symbol("{0}_bit_{1}".format(name, n))
    circuit += cirq.X(q)**bit
  return circuit


class QNN(tf.keras.Model):
  """Operations on parameterized unitaries with bitstring inputs."""

  def __init__(
      self,
      pqc,
      *,
      symbols=None,
      values=None,
      initializer=tf.keras.initializers.RandomUniform(0, 2 * np.pi),
      backend="noiseless",
      differentiator=None,
      is_analytic=False,
      name=None,
  ):
    """Initializes a QNN.

    Args:
      pqc: Representation of a parameterized quantum circuit.
      symbols: Optional 1-D `tf.Tensor` of strings which are the parameters of
        the QNN.  When `None`, parameters are inferred from the given PQC.
      values: Optional 1-D `tf.Tensor` of floats which are the parameter values
        corresponding to the symbols.  When `None`, parameters are chosen via
        `initializer` instead.
      initializer: A "tf.keras.initializers.Initializer" which specifies how to
        initialize the values of the parameters in `circuit`.  This argument is
        ignored if `values` is not None.
      backend: Optional Python `object` that specifies what backend TFQ will use
        for operations involving this QNN. Options are {"noisy", "noiseless"},
        or however users may also specify a preconfigured cirq execution
        object to use instead, which must inherit `cirq.Sampler`.
      differentiator: Either None or a `tfq.differentiators.Differentiator`,
        which specifies how to take the derivative of a quantum circuit.
      is_analytic: bool flag that enables is_analytic methods. If True, then backend
        must also be "noiseless".
      name: Identifier for this QNN.
    """
    super().__init__(name=name)

    if not isinstance(pqc, cirq.Circuit):
      raise TypeError("pqc must be a cirq.Circuit object."
                      " Given: {}".format(pqc))

    if symbols is None:
      raw_symbols = list(sorted(tfq.util.get_circuit_symbols(pqc)))
      symbols = tf.constant([str(x) for x in raw_symbols], dtype=tf.string)
    self._symbols = symbols

    if values is None:
      values = initializer(shape=[tf.shape(self._symbols)[0]])
    self.values = tf.Variable(
        initial_value=values, name=f"{self.name}_pqc_values")

    self._pqc = tfq.convert_to_tensor([pqc])
    self._inverse_pqc = tfq.convert_to_tensor([pqc**-1])

    self._raw_qubits = sorted(pqc.all_qubits())
    self._qubits = util.qubits_to_indices(self._raw_qubits)

    _bit_circuit = bit_circuit(self._raw_qubits)
    bit_symbols = list(sorted(tfq.util.get_circuit_symbols(_bit_circuit)))
    self._bit_symbols = tf.constant([str(x) for x in bit_symbols])
    self._bit_circuit = tfq.convert_to_tensor([_bit_circuit])

    self._differentiator = differentiator
    self._sample_layer = tfq.layers.Sample(backend=backend)
    if backend == "noiseless" or backend is None:
      self._backend = "noiseless"
      self._is_analytic = is_analytic
      self._expectation_layer = tfq.layers.Expectation(
          backend=backend, differentiator=differentiator)
    else:
      self._backend = backend
      self._is_analytic = False
      self._expectation_layer = tfq.layers.SampledExpectation(
          backend=backend, differentiator=differentiator)

    if self.is_analytic:
      self._unitary_layer = tfq.layers.Unitary()

  @property
  def raw_qubits(self):
    return self._raw_qubits

  @property
  def qubits(self):
    return self._qubits

  @property
  def symbols(self):
    return self._symbols

  @property
  def trainable_variables(self):
    return [self.values]

  @trainable_variables.setter
  def trainable_variables(self, value):
    self.values = value[0]

  @property
  def backend(self):
    return self._backend

  @property
  def differentiator(self):
    return self._differentiator

  @property
  def is_analytic(self):
    return self._is_analytic

  def copy(self):
    qnn = QNN(
        tfq.from_tensor(self.pqc(resolve=False))[0],
        backend=self.backend,
        differentiator=self.differentiator,
        is_analytic=self.is_analytic,
        name=self.name)
    qnn.values.assign(self.values)
    return qnn

  def _sample_function(self,
                       circuits,
                       counts,
                       mask=True,
                       reduce=True,
                       unique=True):
    """General function for sampling from circuits."""
    samples = self._sample_layer(
        circuits, repetitions=tf.expand_dims(tf.math.reduce_max(counts), 0))
    if mask:
      num_samples_mask = tf.cast((tf.ragged.range(counts) + 1).to_tensor(),
                                 tf.bool)
      num_samples_mask = tf.map_fn(tf.random.shuffle, num_samples_mask)
      samples = tf.ragged.boolean_mask(samples, num_samples_mask)
    if unique:
      samples = samples.values.to_tensor()
      return util.unique_bitstrings_with_counts(samples)
    elif reduce:
      samples = samples.values.to_tensor()
    return samples

  def _expectation_function(self, circuits, counts, operators, reduce=True):
    """General function for taking sampled expectations from circuits.

    `counts[i]` sets the weight of `circuits[i]` in the expectation.
    Additionally, if `self.is_analytic` is false, `counts[i]` samples are drawn
    from `circuits[i]` and used to compute each expectation in `operators`.
    """
    num_circuits = tf.shape(circuits)[0]
    num_operators = tf.shape(operators)[0]
    tiled_values = tf.tile(tf.expand_dims(self.values, 0), [num_circuits, 1])
    tiled_operators = tf.tile(tf.expand_dims(operators, 0), [num_circuits, 1])
    if self.backend == "noiseless":
      expectations = self._expectation_layer(
          circuits,
          symbol_names=self.symbols,
          symbol_values=tiled_values,
          operators=tiled_operators,
      )
    else:
      expectations = self._expectation_layer(
          circuits,
          symbol_names=self.symbols,
          symbol_values=tiled_values,
          operators=tiled_operators,
          repetitions=tf.tile(tf.expand_dims(counts, 1), [1, num_operators]),
      )
    if reduce:
      probs = tf.cast(counts, tf.float32) / tf.cast(
          tf.reduce_sum(counts), tf.float32)
      return tf.reduce_sum(tf.transpose(probs * tf.transpose(expectations)), 0)
    return expectations

  def pqc(self, resolve=True):
    if resolve:
      return tfq.resolve_parameters(self._pqc, self.symbols,
                                    tf.expand_dims(self.values, 0))
    return self._pqc

  def inverse_pqc(self, resolve=True):
    if resolve:
      return tfq.resolve_parameters(self._inverse_pqc, self.symbols,
                                    tf.expand_dims(self.values, 0))
    return self._inverse_pqc

  def circuits(self, bitstrings, resolve=True):
    """Returns the current circuits for this QNN given bitstrings.

      Args:
        bitstrings: 2D tensor of dtype `tf.int8` whose entries are bits. These
          specify the state inputs to use in the returned set of circuits.
        resolve: bool which says whether or not to resolve the QNN
          unitary before appending to the bit injection circuits.

      Returns:
        1D tensor of strings which represent the current QNN circuits.
      """
    num_bitstrings = tf.shape(bitstrings)[0]
    bit_circuits = tfq.resolve_parameters(
        tf.tile(self._bit_circuit, [num_bitstrings]), self._bit_symbols,
        tf.cast(bitstrings, tf.float32))
    pqcs = tf.tile(self.pqc(resolve=resolve), [num_bitstrings])
    return tfq.append_circuit(bit_circuits, pqcs)

  def sample(self, bitstrings, counts, mask=True, reduce=True, unique=True):
    """Returns bitstring samples from the QNN.

      Args:
        bitstrings: 2D tensor of dtype `tf.int8` whose entries are bits. These
          specify the state inputs to the unitary of this QNN.
        counts: 1D tensor of dtype `tf.int32` such that `counts[i]` is the
          number of samples to draw from `self.pqc|bitstrings[i]>`.

      Returns:
        ragged_samples: `tf.RaggedTensor` of DType `tf.int8` structured such
          that `ragged_samples[i]` contains `counts[i]` bitstrings drawn from
          `self.u|bitstrings[i]>`.
    """
    circuits = self.circuits(bitstrings)
    return self._sample_function(
        circuits, counts, mask=mask, reduce=reduce, unique=unique)

  def expectation(self, bitstrings, counts, operators, reduce=True):
    """Returns the expectation values of the operators against the QNN.

      Args:
        bitstrings: 2D tensor of dtype `tf.int8` whose entries are bits.
        counts: 1D tensor of dtype `tf.int32` such that `counts[i]` is the
          relative weight of `bitstrings[i]` when computing expectations.
        operators: 1D tensor of strings, the result of calling
          `tfq.convert_to_tensor` on a list of cirq.PauliSum, `[op1, op2, ...]`.
          Will be tiled to measure `<opj>_self.u_dagger|circuits[i]>`
          for each i and j.
        reduce: bool flag for whether or not to average over i.

      Returns:
        1-D tensor of floats which are the bitstring-averaged expectation values
        if `reduce` is True, else 2-D tensor of floats which are per-bitstring
        expectation values.
      """
    circuits = self.circuits(bitstrings, resolve=False)
    return self._expectation_function(
        circuits, counts, operators, reduce=reduce)

  def pulled_back_circuits(self, circuits, resolve=True):
    """Returns the pulled back circuits for this QNN given input quantum data.

      Args:
        circuits: 1-D `tf.Tensor` of type `tf.string` which contains
          circuits serialized by `tfq.convert_to_tensor`. These represent pure
          state samples from the data density matrix.
        resolve: bool tensor which says whether or not to resolve the QNN
          inverse unitary before appending to the data circuits.

      Returns:
        1D tensor of strings which represent the pulled back circuits.
      """
    inverse_pqcs = tf.tile(
        self.inverse_pqc(resolve=resolve), [tf.shape(circuits)[0]])
    return tfq.append_circuit(circuits, inverse_pqcs)

  def pulled_back_sample(self,
                         circuits,
                         counts,
                         mask=True,
                         reduce=True,
                         unique=True):
    """Returns samples from the pulled back data distribution.

      The inputs represent the data density matrix. The inverse of `self.u`
      is appended to create the set of circuits representing the pulled back
      data density matrix. Then, the requested number of bitstrings are sampled
      from each circuit.

      Args:
        circuits: 1-D `tf.Tensor` of type `tf.string` which contains
          circuits serialized by `tfq.convert_to_tensor`. These represent pure
          state samples from the data density matrix.
        counts: 1-D `tf.Tensor` of type `tf.int32`, must be the same size as
          `circuits`. Contains the number of samples to draw from each
          input circuit.

      Returns:
        ragged_samples: `tf.RaggedTensor` of DType `tf.int8` structured such
            that `ragged_samples[i]` contains `counts[i]` bitstrings.
      """
    pulled_back_circuits = self.pulled_back_circuits(circuits)
    return self._sample_function(
        pulled_back_circuits, counts, mask=mask, reduce=reduce, unique=unique)

  def pulled_back_expectation(self, circuits, counts, operators, reduce=True):
    """Returns the expectation values for a given pulled-back dataset.

      Args:
        circuits: 1-D `tf.Tensor` of type `tf.string` which contains
          circuits serialized by `tfq.convert_to_tensor`. These represent pure
          state samples from the data density matrix.
        counts: 1D tensor of dtype `tf.int32` such that `counts[i]` is the
          relative weight of `circuits[i]` when computing expectations.
        operators: 1D tensor of strings, the result of calling
          `tfq.convert_to_tensor` on a list of cirq.PauliSum, `[op1, op2, ...]`.
          Will be tiled to measure `<opj>_self.u_dagger|circuits[i]>`
          for each i and j, then averaged over i.

      Returns:
        1-D tensor of floats which are the bitstring-averaged expectation values
        if `reduce` is True, else 2-D tensor of floats which are per-bitstring
        expectation values.
    """
    pulled_back_circuits = self.pulled_back_circuits(circuits, resolve=False)
    return self._expectation_function(
        pulled_back_circuits, counts, operators, reduce=reduce)

  def pqc_unitary(self):
    if self.is_analytic:
      return self._unitary_layer(self.pqc()).to_tensor()[0]
    raise NotImplementedError()

  def __add__(self, other):
    """Returns QNN which is the pqc of `other` appended to the pqc of `self`.

    Note: the backend, differentiator, and is_analytic attributes may be
    different between self and other. The resulting self + other uses such
    attributes from the QNN with the larger number of qubits.
    """
    if len(self.raw_qubits) >= len(other.raw_qubits):
      copy_qnn = self
    else:
      copy_qnn = other
    new_pqc = tfq.append_circuit(self.pqc(False), other.pqc(False))
    new_symbols = tf.concat([self.symbols, other.symbols], 0)
    new_values = tf.concat([self.values, other.values], 0)
    new_qnn = QNN(
        tfq.from_tensor(new_pqc)[0],
        symbols=new_symbols,
        values=new_values,
        backend=copy_qnn.backend,
        differentiator=copy_qnn.differentiator,
        is_analytic=copy_qnn.is_analytic,
        name=f"{self.name}_plus_{other.name}")
    return new_qnn

  def __pow__(self, exponent):
    """QNN raised to a power, only valid for exponent -1, the inverse."""
    if exponent != -1:
      raise ValueError("Only the inverse (exponent == -1) is supported.")
    new_qnn = self.copy()
    old_pqc = new_qnn.pqc(False)
    old_inverse_pqc = new_qnn.inverse_pqc(False)
    new_qnn._pqc = old_inverse_pqc
    new_qnn._inverse_pqc = old_pqc
    return new_qnn



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implementation of general QHBMs in TFQ."""

import tensorflow as tf
import tensorflow_quantum as tfq


class QHBM(tf.keras.Model):
  """Class for Quantum Hamiltonian-Based Model."""

  def __init__(self, ebm, qnn, name=None):
    super().__init__(name=name)
    self._ebm = ebm
    self._qnn = qnn
    if ebm.has_operator:
      self._operator_shards = tfq.convert_to_tensor(
          ebm.operator_shards(qnn.raw_qubits))

  @property
  def ebm(self):
    return self._ebm

  @property
  def qnn(self):
    return self._qnn

  @property
  def trainable_variables(self):
    return self.ebm.trainable_variables + self.qnn.trainable_variables

  @trainable_variables.setter
  def trainable_variables(self, value):
    self.ebm.trainable_variables = value[:len(self.ebm.trainable_variables)]
    self.qnn.trainable_variables = value[len(self.ebm.trainable_variables):]

  @property
  def operator_shards(self):
    if self.ebm.has_operator:
      return self._operator_shards
    raise NotImplementedError()

  @property
  def raw_qubits(self):
    return self.qnn.raw_qubits

  @property
  def qubits(self):
    return self.qnn.qubits

  @property
  def is_analytic(self):
    return self.ebm.is_analytic and self.qnn.is_analytic

  def copy(self):
    return QHBM(self.ebm.copy(), self.qnn.copy(), name=self.name)

  def circuits(self, num_samples):
    bitstrings, counts = self.ebm.sample(num_samples)
    circuits = self.qnn.circuits(bitstrings)
    return circuits, counts

  def sample(self, num_samples, mask=True, reduce=True, unique=True):
    bitstrings, counts = self.ebm.sample(num_samples)
    return self.qnn.sample(
        bitstrings, counts, mask=mask, reduce=reduce, unique=unique)

  def expectation(self, operators, num_samples, reduce=True):
    bitstrings, counts = self.ebm.sample(num_samples)
    return self.qnn.expectation(bitstrings, counts, operators, reduce=reduce)

  def probabilities(self):
    return self.ebm.probabilities()

  def log_partition_function(self):
    return self.ebm.log_partition_function()

  def entropy(self):
    return self.ebm.entropy()

  def unitary_matrix(self):
    return self.qnn.pqc_unitary()

  def density_matrix(self):
    probabilities = tf.cast(self.probabilities(), tf.complex64)
    unitary_matrix = self.unitary_matrix()
    unitary_probs = tf.multiply(
        unitary_matrix,
        tf.tile(
            tf.expand_dims(probabilities, 0), [tf.shape(unitary_matrix)[0], 1]))
    return tf.matmul(unitary_probs, tf.linalg.adjoint(unitary_matrix))

  def fidelity(self, sigma: tf.Tensor):
    """TODO: convert to tf.keras.metric.Metric
    Calculate the fidelity between a QHBM and a density matrix.
        Args:
          sigma: 2-D `tf.Tensor` of dtype `complex64` representing the right
            density matrix in the fidelity calculation.
        Returns:
          A scalar `tf.Tensor` which is the fidelity between the density matrix
            represented by this QHBM and `sigma`.
        """
    e_rho = tf.cast(self.probabilities(), tf.complex128)
    v_rho = tf.cast(self.unitary_matrix(), tf.complex128)
    sqrt_e_rho = tf.sqrt(e_rho)
    v_rho_sqrt_e_rho = tf.multiply(
        v_rho, tf.tile(tf.expand_dims(sqrt_e_rho, 0), (tf.shape(v_rho)[0], 1)))
    rho_sqrt = tf.linalg.matmul(v_rho_sqrt_e_rho, tf.linalg.adjoint(v_rho))
    omega = tf.linalg.matmul(
        tf.linalg.matmul(rho_sqrt, tf.cast(sigma, tf.complex128)), rho_sqrt)
    # TODO(zaqqwerty): find convincing proof that omega is hermitian,
    # in order to go back to eigvalsh.
    e_omega = tf.linalg.eigvals(omega)
    return tf.cast(
        tf.math.abs(tf.math.reduce_sum(tf.math.sqrt(e_omega)))**2, tf.float32)



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for modeling energy functions."""

import abc
from typing import List, Union

import cirq
import tensorflow as tf

from qhbmlib import energy_model_utils


class BitstringEnergy(tf.keras.layers.Layer):
  r"""Class for representing an energy function over bitstrings.

  Keras Layer which can be interpreted as outputting an unnormalized
  log-probability for each given bit-string x, written E(x).  Hence, this class
  implicitly defines a probability distribution over all bit-strings given by
  $$p(x) = \frac{\exp(-E(x))}{\sum_x \exp(-E(x))}.$$

  Moving to its use in QHBMs: each bit-string can also be interpreted as
  an index for an entry of the diagonal eigenvalue matrix in the spectral
  representation of a density operator.  Hence, for a QHBM, inference
  corresponds to sampling computational basis states $|x><x|$ for $x ~ p$,
  where $p$ is the probability distribution written above.
  """

  def __init__(self,
               bits: List[int],
               energy_layers: List[tf.keras.layers.Layer],
               name: Union[None, str] = None):
    """Initializes a BitstringEnergy.

    Args:
      bits: Unique labels for the bits on which this distribution is supported.
      energy_layers: Concatenation of these layers yields trainable map from
        bitstrings to scalars.
      name: Optional name for the model.
    """
    super().__init__(name=name)
    self._bits = energy_model_utils.check_bits(bits)
    self._energy_layers = energy_layers

  @property
  def num_bits(self):
    """Number of bits on which this layer acts."""
    return len(self.bits)

  @property
  def bits(self):
    """Labels for the bits on which this distribution is supported."""
    return self._bits

  @property
  def energy_layers(self):
    """List of keras layers which, when stacked, map bitstrings to energies.

    This list of layers is where the caller would access model weights to be
    updated from a secondary model or hypernetwork.
    """
    return self._energy_layers

  def build(self, input_shape):
    """Builds all the internal layers."""
    x = input_shape
    for layer in self._energy_layers:
      x = layer.compute_output_shape(x)

  def call(self, inputs):
    """Returns the energies corresponding to the input bitstrings."""
    x = inputs
    for layer in self._energy_layers:
      x = layer(x)
    return x


class PauliMixin(abc.ABC):
  """Mixin class to add a Pauli Z representation to  BitstringEnergy."""

  @property
  @abc.abstractmethod
  def post_process(self):
    """List of keras layers.  Concatenation of these layers yields
    the trainable map from the operator shard expectations to a single scalar
    which is the average energy.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def operator_shards(self, qubits: List[cirq.GridQubit]):
    """Parameter independent Pauli Z strings to measure.

    Args:
      qubits: List of cirq.GridQubits. objects to measure.

    Returns:
      List of PauliSum objects whose expectation values are fed to
        `operator_expectation` to compute average energy.
    """
    raise NotImplementedError()

  def operator_expectation(self, expectation_shards: tf.Tensor):
    """Computes the average energy given operator shard expectation values."""
    x = expectation_shards
    for layer in self.post_process:
      x = layer(x)
    return x


class BernoulliEnergy(BitstringEnergy, PauliMixin):
  """Tensor product of coin flip distributions.

  Note that we parameterize using the energy of a spin in a magnetic field,
  which is offset from the log probability of a typical Bernoulli.
  """

  def __init__(self,
               bits: List[int],
               initializer: tf.keras.initializers.Initializer = tf.keras
               .initializers.RandomUniform(),
               name: Union[None, str] = None):
    """Initializes a BernoulliEnergy.

    Args:
      bits: Unique labels for the bits on which this distribution is supported.
      initializer: A `tf.keras.initializers.Initializer` which specifies how to
        initialize the values of the parameters.
      name: Optional name for the model.
    """
    pre_process = [energy_model_utils.SpinsFromBitstrings()]
    post_process = [energy_model_utils.VariableDot(initializer=initializer)]
    super().__init__(bits, pre_process + post_process, name)
    self._post_process = post_process

  @property
  def logits(self):
    r"""Returns the current logits of the distribution.

    For our Bernoulli distribution, let $p$ be the probability of bit being `1`.
    In this case, we have $p = \frac{e^{theta}}{{e^{theta}+e^{-theta}}}$.
    Therefore, each independent logit is:
    $$logit = \log\frac{p}{1-p} = \log\frac{e^{theta}}{e^{-theta}}
            = \log{e^{2*theta}} = 2*theta$$
    """
    return 2 * self.post_process[0].kernel

  @property
  def post_process(self):
    """See base class description."""
    return self._post_process

  def operator_shards(self, qubits):
    """See base class description."""
    return [cirq.PauliSum.from_pauli_strings(cirq.Z(q)) for q in qubits]


class KOBE(BitstringEnergy, PauliMixin):
  """Kth Order Binary Energy function."""

  def __init__(self,
               bits: List[int],
               order: int,
               initializer: tf.keras.initializers.Initializer = tf.keras
               .initializers.RandomUniform(),
               name: Union[None, str] = None):
    """Initializes a KOBE.

    Args:
      bits: Each entry is an index on which the distribution is supported.
      order: The order of the KOBE.
      initializer: Specifies how to initialize the values of the parameters.
      name: Optional name for the model.
    """
    parity_layer = energy_model_utils.Parity(bits, order)
    self._num_terms = parity_layer.num_terms
    self._indices = parity_layer.indices
    pre_process = [energy_model_utils.SpinsFromBitstrings(), parity_layer]
    post_process = [energy_model_utils.VariableDot(initializer=initializer)]
    super().__init__(bits, pre_process + post_process, name)
    self._post_process = post_process

  @property
  def post_process(self):
    """See base class description."""
    return self._post_process

  def operator_shards(self, qubits: List[cirq.GridQubit]):
    """See base class description."""
    ops = []
    for i in range(self._num_terms):
      string_factors = []
      for loc in self._indices[i]:
        string_factors.append(cirq.Z(qubits[loc]))
        string = cirq.PauliString(string_factors)
      ops.append(cirq.PauliSum.from_pauli_strings(string))
    return ops



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qhbmlib']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow-probability==0.12.0',
 'tensorflow-quantum==0.5.1',
 'tensorflow==2.4.1']

setup_kwargs = {
    'name': 'qhbmlib',
    'version': '0.2.2.dev20220115030238',
    'description': 'Quantum Hamiltonian-Based Models built on TensorFlow Quantum',
    'long_description': '# QHBM Library\n\nThis repository is a collection of tools for building and training\nQuantum Hamiltonian-Based Models.  These tools depend on\n[TensorFlow Quantum](https://www.tensorflow.org/quantum),\nand are thus compatible with both real and simulated quantum computers.\n\n[Installation instructions](https://github.com/google/qhbm-library/blob/main/docs/INSTALL.md) and [contribution instructions](https://github.com/google/qhbm-library/blob/main/docs/contributing.md) can be found in the docs folder.\n\nThis is not an officially supported Google product.\n',
    'author': 'The QHBM Library Authors',
    'author_email': 'no-reply@google.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/google/qhbm-library',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Import QHBM library modules.*"""

from qhbmlib import architectures
from qhbmlib import ebm
from qhbmlib import hamiltonian
from qhbmlib import qhbm
from qhbmlib import qmhl
from qhbmlib import qnn
from qhbmlib import util
from qhbmlib import vqt



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Impementations of the VQT loss and its derivatives."""

import tensorflow as tf


def vqt(model, num_samples, hamiltonian, beta):
  """Computes the VQT loss of a given QHBM against given thermal state params.

  This function is differentiable within a `tf.GradientTape` scope.

  Args:
    model: A `qhbm.QHBM` which is the model whose loss is to be calculated.
    num_samples: A scalar `tf.Tensor` specifying the number of samples to draw
      from the EBM of `model` when estimating the loss and its gradients.
    hamiltonian: 1D tensor of strings with one entry, the result of calling
      `tfq.convert_to_tensor` on a list containing one cirq.PauliSum, `[op]`.
      Here, `op` is the Hamiltonian against which the loss is calculated.
    beta: A scalar `tf.Tensor` which is the inverse temperature at which the
      loss is calculated.

  Returns:
    The VQT loss.
  """

  @tf.custom_gradient
  def loss(trainable_variables):
    # We use `model.qnn.trainable_variables` / `model.ebm.trainable_variables`
    # instead
    del trainable_variables
    bitstrings, counts = model.ebm.sample(num_samples)
    probs = tf.cast(counts, tf.float32) / tf.cast(num_samples, tf.float32)
    expectation = tf.squeeze(
        model.qnn.expectation(bitstrings, counts, hamiltonian), -1)
    if model.is_analytic:
      entropy = model.entropy()
    else:
      entropy = -tf.reduce_sum(probs * tf.math.log(probs))

    def grad(grad_y, variables=None):
      with tf.GradientTape() as tape:
        tape.watch(model.qnn.trainable_variables)
        beta_expectations = beta * tf.squeeze(
            model.qnn.expectation(
                bitstrings, counts, hamiltonian, reduce=False), -1)
        beta_expectation = tf.reduce_sum(probs * beta_expectations)
      grad_qnn = tape.gradient(beta_expectation, model.qnn.trainable_variables)
      grad_qnn = [grad_y * g for g in grad_qnn]

      with tf.GradientTape() as tape:
        tape.watch(model.ebm.trainable_variables)
        energies = model.ebm.energy(bitstrings)
      energy_jac = tape.jacobian(energies, model.ebm.trainable_variables)
      probs_diffs = probs * (beta_expectations - energies)
      avg_diff = tf.reduce_sum(probs_diffs)
      grad_ebm = [
          grad_y *
          (avg_diff * tf.reduce_sum(tf.transpose(probs * tf.transpose(g)), 0) -
           tf.reduce_sum(tf.transpose(probs_diffs * tf.transpose(g)), 0))
          for g in energy_jac
      ]
      grad_qhbm = grad_ebm + grad_qnn
      if variables is None:
        return grad_qhbm
      return grad_qhbm, [tf.zeros_like(g) for g in grad_qhbm]

    return beta * expectation - entropy, grad

  return loss(model.trainable_variables)



# pylint: skip-file
# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Module for defining and sampling from EBMs."""

import abc
import collections
import itertools

import cirq
import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_quantum as tfq

from qhbmlib import util

from tensorflow_probability.python.mcmc.internal import util as mcmc_util


def probability_to_logit(probability):
  p = tf.cast(probability, tf.dtypes.float32)
  return tf.math.log(p) - tf.math.log(1 - p)


def logit_to_probability(logit_in):
  logit = tf.cast(logit_in, tf.dtypes.float32)
  return tf.math.divide(tf.math.exp(logit), 1 + tf.math.exp(logit))


class EnergyFunction(tf.keras.Model, abc.ABC):

  def __init__(self, name=None):
    super().__init__(name=name)

  @property
  @abc.abstractmethod
  def num_bits(self):
    raise NotImplementedError()

  @property
  def trainable_variables(self):
    return super().trainable_variables

  @trainable_variables.setter
  @abc.abstractmethod
  def trainable_variables(self, value):
    raise NotImplementedError()

  @property
  def has_operator(self):
    return False

  @abc.abstractmethod
  def energy(self, bitstrings):
    raise NotImplementedError()

  def operator_shards(self, qubits):
    raise NotImplementedError()

  def operator_expectation(self, expectation_shards):
    raise NotImplementedError()


class KOBE(EnergyFunction):

  def __init__(self,
               num_bits,
               order,
               initializer=tf.keras.initializers.RandomUniform(),
               name=None):
    super().__init__(name=name)
    self._num_bits = num_bits
    self._order = order
    indices_list = []
    for i in range(1, order + 1):
      combos = itertools.combinations(range(num_bits), i)
      indices_list.extend(list(combos))
    self._indices = tf.ragged.stack(indices_list)
    self._num_terms = tf.constant(len(indices_list))
    self.kernel = self.add_weight(
        name='kernel',
        shape=[self._num_terms],
        initializer=initializer,
        trainable=True)

  @property
  def trainable_variables(self):
    return [self.kernel]

  @trainable_variables.setter
  def trainable_variables(self, value):
    self.kernel = value[0]

  @property
  def num_bits(self):
    return self._num_bits

  @property
  def has_operator(self):
    return True

  @property
  def order(self):
    return self._order

  def copy(self):
    kobe = KOBE(self.num_bits, self.order, name=self.name)
    kobe.kernel.assign(self.kernel)
    return kobe

  def energy(self, bitstrings):
    spins = 1 - 2 * bitstrings
    parities_t = tf.zeros(
        [self._num_terms, tf.shape(bitstrings)[0]], dtype=tf.float32)
    for i in tf.range(self._num_terms):
      parity = tf.reduce_prod(tf.gather(spins, self._indices[i], axis=-1), -1)
      parities_t = tf.tensor_scatter_nd_update(parities_t, [[i]], [parity])
    return tf.reduce_sum(tf.transpose(parities_t) * self.kernel, -1)

  def operator_shards(self, qubits):
    ops = []
    for i in range(self._num_terms):
      string_factors = []
      for loc in self._indices[i]:
        string_factors.append(cirq.Z(qubits[loc]))
      string = cirq.PauliString(string_factors)
      ops.append(cirq.PauliSum.from_pauli_strings(string))
    return ops

  def operator_expectation(self, expectation_shards):
    return tf.reduce_sum(expectation_shards * self.kernel, -1)


class MLP(EnergyFunction):

  def __init__(self, num_bits, units, activations, name=None):
    super().__init__(name=name)
    self._num_bits = num_bits
    self._hidden_layers = [
        tf.keras.layers.Dense(u, activation=a)
        for u, a in zip(units, activations)
    ]
    self._energy_layer = tf.keras.layers.Dense(1)
    self.build([1, num_bits])

  @property
  def num_bits(self):
    return self._num_bits

  @property
  def trainable_variables(self):
    trainable_variables = []
    for layer in self.layers:
      trainable_variables.extend([layer.kernel, layer.bias])
    return trainable_variables

  @trainable_variables.setter
  def trainable_variables(self, value):
    i = 0
    for layer in self.layers:
      layer.kernel = value[i]
      layer.bias = value[i + 1]
      i += 2

  def copy(self):
    mlp = tf.keras.models.clone_model(self)
    for i in tf.range(len(mlp.trainable_variables)):
      mlp.trainable_variables[i].assign(self.trainable_variables[i])
    return mlp

  def call(self, bitstrings):
    x = bitstrings
    for hidden_layer in self._hidden_layers:
      x = hidden_layer(x)
    x = self._energy_layer(x)
    return tf.squeeze(x, -1)

  def energy(self, bitstrings):
    return self(bitstrings)


class EnergySampler(abc.ABC):

  @property
  @abc.abstractmethod
  def energy_function(self):
    raise NotImplementedError()

  @abc.abstractmethod
  def sample(self, num_samples, unique=True):
    raise NotImplementedError()


class EnergyKernel(tfp.mcmc.TransitionKernel, abc.ABC):

  @property
  @abc.abstractmethod
  def energy_function(self):
    raise NotImplementedError()


class UncalibratedGWGResults(
    mcmc_util.PrettyNamedTupleMixin,
    collections.namedtuple('UncalibratedGWGResults',
                           ['target_log_prob', 'log_acceptance_correction'])):
  __slots__ = ()


class UncalibratedGWG(EnergyKernel):

  def __init__(self,
               energy_function,
               gradient=True,
               temperature=2.0,
               num_samples=1,
               name=None):
    self._parameters = dict(
        energy_function=energy_function,
        gradient=gradient,
        temperature=temperature,
        num_samples=num_samples,
        name=name)
    self._diff_function = self._grad_diff_function if gradient else self._exact_diff_function

  @property
  def energy_function(self):
    return self._parameters['energy_function']

  @property
  def gradient(self):
    return self._parameters['gradient']

  @property
  def temperature(self):
    return self._parameters['temperature']

  @property
  def num_samples(self):
    return self._parameters['num_samples']

  @property
  def name(self):
    return self._parameters['name']

  @property
  def parameters(self):
    return self._parameters

  @property
  def is_calibrated(self):
    return False

  def copy(self):
    return UncalibratedGWG(
        self.energy_function.copy(),
        gradient=self.gradient,
        temperature=self.temperature,
        num_samples=self.num_samples,
        name=self.name)

  def _exact_diff_function(self, current_state):
    current_state_t = tf.transpose(current_state)
    diff_t = tf.zeros_like(current_state_t, dtype=tf.float32)
    current_energy = self.energy_function.energy(current_state)

    for i in tf.range(tf.shape(current_state)[-1]):
      pert_state = tf.transpose(
          tf.tensor_scatter_nd_update(current_state_t, [[i]],
                                      tf.expand_dims(1 - current_state_t[i],
                                                     0)))
      diff_t = tf.tensor_scatter_nd_update(
          diff_t, [[i]],
          tf.expand_dims(
              current_energy - self.energy_function.energy(pert_state), 0))
    return tf.transpose(diff_t)

  def _grad_diff_function(self, current_state):
    current_state = tf.cast(current_state, tf.float32)
    with tf.GradientTape() as tape:
      tape.watch(current_state)
      energy = self.energy_function.energy(current_state)
    grad = tape.gradient(energy, current_state)
    return (2 * current_state - 1) * grad

  def one_step(self, current_state, previous_kernel_results):
    forward_diff = self._diff_function(current_state)
    forward_dist = tfp.distributions.OneHotCategorical(
        logits=forward_diff / self.temperature, dtype=tf.int8)
    all_changes = forward_dist.sample(self.num_samples)
    total_change = tf.cast(tf.reduce_sum(all_changes, 0) > 0, tf.int8)
    next_state = tf.bitwise.bitwise_xor(total_change, current_state)

    target_log_prob = -1 * self.energy_function.energy(next_state)

    forward_log_prob = tf.reduce_sum(forward_dist.log_prob(all_changes), 0)
    backward_diff = self._diff_function(next_state)
    backward_dist = tfp.distributions.OneHotCategorical(
        logits=backward_diff / self.temperature, dtype=tf.int8)
    backward_log_prob = tf.reduce_sum(backward_dist.log_prob(all_changes), 0)
    log_acceptance_correction = backward_log_prob - forward_log_prob

    kernel_results = UncalibratedGWGResults(
        target_log_prob=target_log_prob,
        log_acceptance_correction=log_acceptance_correction)

    return next_state, kernel_results

  def bootstrap_results(self, init_state):
    target_log_prob = -1 * self.energy_function.energy(init_state)
    kernel_results = UncalibratedGWGResults(
        target_log_prob=target_log_prob,
        log_acceptance_correction=tf.zeros_like(target_log_prob))
    return kernel_results


class MetropolisHastings(tfp.mcmc.MetropolisHastings, EnergyKernel):

  def __init__(self, inner_kernel, name=None):
    super().__init__(inner_kernel, name=name)

  @property
  def energy_function(self):
    return self.inner_kernel.energy_function

  def copy(self):
    return MetropolisHastings(self.inner_kernel.copy())


class GWG(MetropolisHastings):

  def __init__(self,
               energy_function,
               gradient=True,
               temperature=2.0,
               num_samples=1,
               name=None):
    self._impl = MetropolisHastings(
        UncalibratedGWG(
            energy_function,
            gradient=gradient,
            temperature=temperature,
            num_samples=num_samples,
            name=name))
    self._parameters = self._impl.inner_kernel.parameters.copy()

  @property
  def energy_function(self):
    return self._impl.inner_kernel.energy_function

  @property
  def gradient(self):
    return self._impl.inner_kernel.gradient

  @property
  def temperature(self):
    return self._impl.inner_kernel.temperature

  @property
  def num_samples(self):
    return self._impl.inner_kernel.num_samples

  @property
  def name(self):
    return self._impl.inner_kernel.name

  @property
  def parameters(self):
    return self._impl.inner_kernel.parameters

  @property
  def is_calibrated(self):
    return True

  def copy(self):
    return GWG(
        self.energy_function.copy(),
        gradient=self.gradient,
        temperature=self.temperature,
        num_samples=self.num_samples,
        name=self.name)

  def one_step(self, current_state, previous_kernel_results):
    return self._impl.one_step(current_state, previous_kernel_results)

  def bootstrap_results(self, init_state):
    return self._impl.bootstrap_results(init_state)


class MCMC(EnergySampler):

  def __init__(self,
               kernel,
               num_chains=1,
               buffer_capacity=1000,
               buffer_probability=1,
               num_burnin_steps=0,
               num_steps_between_results=0,
               parallel_iterations=10,
               name=None):
    self._kernel = kernel
    self._num_chains = num_chains
    self._num_bits = kernel.energy_function.num_bits
    self._buffer_capacity = buffer_capacity
    self._buffer_probability = buffer_probability
    self._num_burnin_steps = num_burnin_steps
    self._num_steps_between_results = num_steps_between_results
    self._parallel_iterations = parallel_iterations
    self._name = name

    self._buffer = tf.queue.RandomShuffleQueue(
        buffer_capacity, 0, [tf.int8], shapes=[self._num_bits])

  @property
  def energy_function(self):
    return self.kernel.energy_function

  @property
  def kernel(self):
    return self._kernel

  @property
  def num_chains(self):
    return self._num_chains

  @property
  def buffer_capacity(self):
    return self._buffer_capacity

  @property
  def buffer_probability(self):
    return self._buffer_probability

  @property
  def num_burnin_steps(self):
    return self._num_burnin_steps

  @property
  def num_steps_between_results(self):
    return self._num_steps_between_results

  @property
  def parallel_iterations(self):
    return self._parallel_iterations

  @property
  def name(self):
    return self._name

  def copy(self):
    mcmc = MCMC(
        self.kernel.copy(),
        num_chains=self.num_chains,
        buffer_capacity=self.buffer_capacity,
        buffer_probability=self.buffer_probability,
        num_burnin_steps=self.num_burnin_steps,
        num_steps_between_results=self.num_steps_between_results,
        parallel_iterations=self.parallel_iterations,
        name=self.name)
    mcmc._buffer = tf.queue.QueueBase.from_list(tf.constant(0), [self._buffer])
    return mcmc

  def sample(self, num_samples, unique=True):
    num_results = tf.cast(tf.math.ceil(num_samples / self.num_chains), tf.int32)

    if tf.random.uniform(()) > self.buffer_probability:
      init_state = tf.cast(
          tf.random.uniform([self.num_chains, self._num_bits],
                            maxval=2,
                            dtype=tf.int32), tf.int8)
    else:
      init_state = self._buffer.dequeue_many(
          tf.math.minimum(self.num_chains, self._buffer.size()))
      init_state = tf.concat([
          init_state,
          tf.cast(
              tf.random.uniform(
                  [self.num_chains - tf.shape(init_state)[0], self._num_bits],
                  maxval=2,
                  dtype=tf.int32), tf.int8)
      ], 0)

    previous_kernel_results = self.kernel.bootstrap_results(init_state)

    samples = tfp.mcmc.sample_chain(
        num_results,
        init_state,
        previous_kernel_results=previous_kernel_results,
        kernel=self.kernel,
        num_burnin_steps=self.num_burnin_steps,
        num_steps_between_results=self.num_steps_between_results,
        trace_fn=None,
        parallel_iterations=self.parallel_iterations,
        name=self.name)

    sampled_states = tf.reshape(samples, [-1, self._num_bits])
    if tf.shape(sampled_states)[0] > self.buffer_capacity - self._buffer.size():
      self._buffer.dequeue_many(
          tf.math.minimum(
              tf.shape(sampled_states)[0] -
              (self.buffer_capacity - self._buffer.size()),
              self._buffer.size()))
    self._buffer.enqueue_many(sampled_states[:tf.math.minimum(
        tf.shape(sampled_states)[0], self.buffer_capacity)])

    sampled_states = sampled_states[:num_samples]
    if unique:
      return util.unique_bitstrings_with_counts(sampled_states)
    return sampled_states


class EBM(tf.keras.Model):

  def __init__(self,
               energy_function,
               energy_sampler,
               is_analytic=False,
               name=None):
    super().__init__(name=name)
    self._energy_function = energy_function
    self._energy_sampler = energy_sampler
    self._is_analytic = is_analytic
    if is_analytic:
      self._all_bitstrings = tf.constant(
          list(itertools.product([0, 1], repeat=energy_function.num_bits)),
          dtype=tf.int8)

  @property
  def num_bits(self):
    return self._energy_function.num_bits

  @property
  def has_operator(self):
    return self._energy_function.has_operator

  @property
  def is_analytic(self):
    return self._is_analytic

  def copy(self):
    if self._energy_sampler is not None:
      energy_sampler = self._energy_sampler.copy()
      energy_function = energy_sampler.energy_function
    else:
      energy_sampler = None
      energy_function = self._energy_function.copy()
    return EBM(
        energy_function,
        energy_sampler,
        is_analytic=self.is_analytic,
        name=self.name)

  def energy(self, bitstrings):
    return self._energy_function.energy(bitstrings)

  def operator_shards(self, qubits):
    return self._energy_function.operator_shards(qubits)

  def operator_expectation(self, expectation_shards):
    return self._energy_function.operator_expectation(expectation_shards)

  def sample(self, num_samples, unique=True):
    if self.is_analytic and self._energy_sampler is None:
      samples = tf.gather(
          self._all_bitstrings,
          tfp.distributions.Categorical(logits=-1 *
                                        self.energies()).sample(num_samples))
      if unique:
        return util.unique_bitstrings_with_counts(samples)
      return samples
    else:
      return self._energy_sampler.sample(num_samples, unique=unique)

  def energies(self):
    if self.is_analytic:
      return self.energy(self._all_bitstrings)
    raise NotImplementedError()

  def probabilities(self):
    if self.is_analytic:
      return tf.exp(-self.ebm.energies()) / tf.exp(
          self.log_partition_function())
    raise NotImplementedError()

  def log_partition_function(self):
    if self.is_analytic:
      return tf.reduce_logsumexp(-1 * self.energies())
    raise NotImplementedError()

  def entropy(self):
    if self.is_analytic:
      return tfp.distributions.Categorical(logits=-1 *
                                           self.energies()).entropy()
    raise NotImplementedError()


class Bernoulli(EBM):

  def __init__(self,
               num_bits,
               initializer=tf.keras.initializers.RandomUniform(),
               is_analytic=False,
               name=None):
    tf.keras.Model.__init__(self, name=name)
    self._num_bits = num_bits
    self.kernel = self.add_weight(
        name=f'kernel',
        shape=[self.num_bits],
        initializer=initializer,
        trainable=True)
    self._is_analytic = is_analytic
    if is_analytic:
      self._all_bitstrings = tf.constant(
          list(itertools.product([0, 1], repeat=num_bits)), dtype=tf.int8)

  @property
  def num_bits(self):
    return self._num_bits

  @property
  def trainable_variables(self):
    return [self.kernel]

  @trainable_variables.setter
  def trainable_variables(self, value):
    self.kernel = value[0]

  @property
  def has_operator(self):
    return True

  @property
  def is_analytic(self):
    return self._is_analytic

  def copy(self):
    bernoulli = Bernoulli(
        self.num_bits, is_analytic=self.is_analytic, name=self.name)
    bernoulli.kernel.assign(self.kernel)
    return bernoulli

  def energy(self, bitstrings):
    return tf.reduce_sum(
        tf.cast(1 - 2 * bitstrings, tf.float32) * self.kernel, -1)

  def operator_shards(self, qubits):
    return [
        cirq.PauliSum.from_pauli_strings(cirq.Z(qubits[i]))
        for i in range(self.num_bits)
    ]

  def operator_expectation(self, expectation_shards):
    return tf.reduce_sum(expectation_shards * self.kernel, -1)

  def sample(self, num_samples, unique=True):
    r"""Fairly samples from the EBM defined by `energy`.

        For Bernoulli distribution, let $p$ be the probability of bit being `1`.
        In this case, we have $p = \frac{e^{theta}}{{e^{theta}+e^{-theta}}}$.
        Therefore, each independent logit is:
          $$logit = \log\frac{p}{1-p} = \log\frac{e^{theta}}{e^{-theta}}
                 = \log{e^{2*theta}} = 2*theta$$
        """
    samples = tfp.distributions.Bernoulli(
        logits=2 * self.kernel, dtype=tf.int8).sample(num_samples)
    if unique:
      return util.unique_bitstrings_with_counts(samples)
    return samples

  def energies(self):
    if self.is_analytic:
      return self.energy(self._all_bitstrings)
    raise NotImplementedError()

  def probabilities(self):
    if self.is_analytic:
      return tf.exp(-self.energies()) / tf.exp(self.log_partition_function())
    raise NotImplementedError()

  def log_partition_function(self):
    if self.is_analytic:
      return tf.reduce_logsumexp(-1 * self.energies())
    raise NotImplementedError()

  def entropy(self):
    return tf.reduce_sum(
        tfp.distributions.Bernoulli(logits=2 * self.kernel).entropy())



# Copyright 2021 The QHBM Library Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tools for defining quantum circuit models."""

from typing import List, Union

import cirq
import numpy as np
import tensorflow as tf
import tensorflow_quantum as tfq

from qhbmlib import circuit_model_utils


class QuantumCircuit(tf.keras.layers.Layer):
  """Class for representing a quantum circuit."""

  def __init__(self,
               pqc: tf.Tensor,
               qubits: List[cirq.GridQubit],
               symbol_names: tf.Tensor,
               value_layers_inputs: List[Union[tf.Variable, List[tf.Variable]]],
               value_layers: List[List[tf.keras.layers.Layer]],
               name: Union[None, str] = None):
    """Initializes a QuantumCircuit.

    Args:
      pqc: TFQ string representation of a parameterized quantum circuit.
      qubits: The qubits on which `pqc` acts.
      symbol_names: Strings which are used to specify the order in which the
        values in `self.symbol_values` should be placed inside of the circuit.
      value_layers_inputs: Inputs to the `value_layers` argument.
      value_layers: The concatenation of the layers in entry `i` yields a
        trainable map from `value_layers_inputs[i]` to the `i` entry in the list
        of intermediate values.  The list of intermediate values is concatenated
        to yield the values to substitute into the circuit.
      name: Optional name for the model.
    """
    super().__init__(name=name)

    self._pqc = pqc
    self._qubits = sorted(qubits)
    self._symbol_names = symbol_names
    self._value_layers = value_layers
    self._value_layers_inputs = value_layers_inputs

    raw_bit_circuit = circuit_model_utils.bit_circuit(self.qubits)
    bit_symbol_names = list(
        sorted(tfq.util.get_circuit_symbols(raw_bit_circuit)))
    self._bit_symbol_names = tf.constant([str(x) for x in bit_symbol_names])
    self._bit_circuit = tfq.convert_to_tensor([raw_bit_circuit])

  @property
  def qubits(self):
    """Sorted list of the qubits on which this circuit acts."""
    return self._qubits

  @property
  def symbol_names(self):
    """1D tensor of strings which are the free parameters of the circuit."""
    return self._symbol_names

  @property
  def value_layers_inputs(self):
    """List of lists of variables which are inputs to `value_layers`.

    This property (and `value_layers`) is where the caller would access model
    weights to be updated from a secondary model or hypernetwork.
    """
    return self._value_layers_inputs

  @property
  def value_layers(self):
    """List of lists of Keras layers which calculate current parameter values.

    This property (and `value_layers_inputs`) is where the caller would access
    model weights to be updated from a secondary model or hypernetwork.
    """
    return self._value_layers

  @property
  def symbol_values(self):
    """1D `tf.Tensor` of floats specifying the current values of the parameters.

    This should be structured such that `self.symbol_values[i]` is the current
    value of `self.symbol_names[i]` in `self.pqc` and `self.inverse_pqc`.
    """
    # TODO(#123): empty value because concat requires at least two entries.
    intermediate_values = [[]]
    for inputs, layers in zip(self.value_layers_inputs, self.value_layers):
      x = inputs
      for layer in layers:
        x = layer(x)
      intermediate_values.append(x)
    return tf.concat(intermediate_values, 0)

  @property
  def pqc(self):
    """TFQ tensor representation of the parameterized unitary circuit."""
    return self._pqc

  def build(self, input_shape):
    """Builds the layers which calculate the values.

    `input_shape` is unused because it is known to be the shape of
    `self._value_layers_inputs`.
    """
    del input_shape
    for inputs, layers in zip(self.value_layers_inputs, self.value_layers):
      if isinstance(inputs, tf.Variable):
        x = inputs.get_shape()
      else:
        x = [v.get_shape() for v in inputs]
      for layer in layers:
        x = layer.compute_output_shape(x)

  def call(self, inputs):
    """Inputs are bitstrings prepended as initial states to `self.pqc`."""
    num_bitstrings = tf.shape(inputs)[0]
    bit_circuits = tfq.resolve_parameters(
        tf.tile(self._bit_circuit, [num_bitstrings]), self._bit_symbol_names,
        tf.cast(inputs, tf.float32))
    pqcs = tf.tile(self.pqc, [num_bitstrings])
    return tfq.append_circuit(bit_circuits, pqcs)

  def __add__(self, other: "QuantumCircuit"):
    """Returns a QuantumCircuit with `self.pqc` appended to `other.pqc`.

    Note that no new `tf.Variable`s are created, the new QuantumCircuit contains
    the variables in both `self` and `other`.
    """
    if isinstance(other, QuantumCircuit):
      intersection = tf.sets.intersection(
          tf.expand_dims(self.symbol_names, 0),
          tf.expand_dims(other.symbol_names, 0))
      tf.debugging.assert_equal(
          tf.size(intersection.values),
          0,
          message="Circuits to be summed must not have symbols in common.")
      new_pqc = tfq.append_circuit(self.pqc, other.pqc)
      new_qubits = list(set(self.qubits + other.qubits))
      new_symbol_names = tf.concat([self.symbol_names, other.symbol_names], 0)
      new_value_layers_inputs = (
          self.value_layers_inputs + other.value_layers_inputs)
      new_value_layers = self.value_layers + other.value_layers
      new_name = self.name + "_" + other.name
      return QuantumCircuit(new_pqc, new_qubits, new_symbol_names,
                            new_value_layers_inputs, new_value_layers, new_name)
    else:
      raise TypeError

  def __pow__(self, exponent):
    """Returns a QuantumCircuit with inverted `self.pqc`.

    Note that no new `tf.Variable`s are created, the new QuantumCircuit contains
    the same variables as `self`.
    """
    if exponent == -1:
      new_pqc = tfq.from_tensor(self.pqc)[0]**-1
      new_name = self.name + "_inverse"
      return QuantumCircuit(
          tfq.convert_to_tensor([new_pqc]), new_pqc.all_qubits(),
          self.symbol_names, self.value_layers_inputs, self.value_layers,
          new_name)
    else:
      raise ValueError("Only the inverse (exponent == -1) is supported.")


class DirectQuantumCircuit(QuantumCircuit):
  """QuantumCircuit with direct map from model variables to circuit params."""

  def __init__(
      self,
      pqc: cirq.Circuit,
      initializer: tf.keras.initializers.Initializer = tf.keras.initializers
      .RandomUniform(0, 2 * np.pi),
      name: Union[None, str] = None,
  ):
    """Initializes a DirectQuantumCircuit.

    Args:
      pqc: Representation of a parameterized quantum circuit.
      initializer: A `tf.keras.initializers.Initializer` which specifies how to
        initialize the values of the parameters in `circuit`.
      name: Optional name for the model.
    """
    raw_symbol_names = list(sorted(tfq.util.get_circuit_symbols(pqc)))
    symbol_names = tf.constant([str(x) for x in raw_symbol_names],
                               dtype=tf.string)
    values = [tf.Variable(initializer(shape=[len(raw_symbol_names)]))]
    value_layers = [[]]
    super().__init__(
        tfq.convert_to_tensor([pqc]), pqc.all_qubits(), symbol_names, values,
        value_layers)


class QAIA(QuantumCircuit):
  """Quantum circuit defined by a classical energy and a Hamiltonian.

  This circuit model is intended for use with VQT.
  """

  def __init__(self,
               quantum_h_terms: List[cirq.PauliSum],
               classical_h_terms: List[cirq.PauliSum],
               num_layers: int,
               initializer=tf.keras.initializers.RandomUniform(0, 2 * np.pi),
               name=None):
    r"""Initializes a QAIA.

    The ansatz is QAOA-like, with the exponential of the EBM ansatz in place
    of the usual "problem Hamiltonian". Mathematically, it is represented as:

    $$\prod_{\ell=1}^P \left[
        \left(
          \prod_{\bm{b} \in \mathcal{B}_K}
            e^{i\eta_\ell \theta_{\bm{b}}\bm{\hat{Z}}^{\bm{b}}}
        \right)\left(
          \prod_{r\in \mathcal{I}}
            e^{i\gamma_{r\ell}\hat{H}_r}
        \right)
      \right],$$

    where $\hat{H}_r$ is `quantum_h_terms`, $\bm{\hat{Z}}^{\bm{b}}$ is
    `classical_h_terms`, and $P$ is `num_layers`.

    # TODO(#119): add link to new version of the paper.
    For further discussion, see the section "Physics-Inspired Architecture:
    Quantum Adiabatic-Inspired Ansatz" in the QHBM paper.

    Args:
      quantum_h_terms: Non-commuting terms of the target thermal state
        assumed in the QAIA ansatz.
      classical_h_terms: Hamiltonian representation of the EBM chosen to model
        the target thermal state.
      num_layers: How many layers of the ansatz to apply.
      initializer: A `tf.keras.initializers.Initializer` which specifies how to
        initialize the values of the parameters in `circuit`.
      name: Optional name for the model.
    """
    quantum_symbols = []
    classical_symbols = []
    for j in range(num_layers):
      quantum_symbols.append([])
      classical_symbols.append([])
      for k, _ in enumerate(quantum_h_terms):
        quantum_symbols[-1].append(f"gamma_{j}_{k}")
      for k, _ in enumerate(classical_h_terms):
        classical_symbols[-1].append(f"eta_{j}_{k}")

    pqc = cirq.Circuit()
    flat_symbols = []
    for q_symb, c_symb in zip(quantum_symbols, classical_symbols):
      pqc += tfq.util.exponential(quantum_h_terms, coefficients=q_symb)
      pqc += tfq.util.exponential(classical_h_terms, coefficients=c_symb)
      flat_symbols.extend(q_symb + c_symb)
    symbol_names = tf.constant(flat_symbols)

    value_layers_inputs = [[
        tf.Variable(initializer(shape=[num_layers])),  # true etas
        tf.Variable(initializer(shape=[len(classical_h_terms)])),  # thetas
        tf.Variable(
            initializer(shape=[num_layers, len(quantum_h_terms)])),  # gammas
    ]]

    def embed_params(inputs):
      """Tiles up the variables to properly tie QAIA parameters."""
      exp_etas = tf.expand_dims(inputs[0], 1)
      tiled_thetas = tf.tile(
          tf.expand_dims(inputs[1], 0), [tf.shape(inputs[0])[0], 1])
      classical_params = exp_etas * tiled_thetas
      return tf.reshape(tf.concat([classical_params, inputs[2]], 1), [-1])

    value_layers = [[tf.keras.layers.Lambda(embed_params)]]

    super().__init__(
        tfq.convert_to_tensor([pqc]), pqc.all_qubits(), symbol_names,
        value_layers_inputs, value_layers)
